import random, time, collections
from operator import add
import examples.CRFs.crf_model as crf_model
from math import *
######################
# Thank you, Python
neg_inf = -1e1000
inf     =  1e1000

#########################
# Some ExpSum
def exp_sum(z):
    cdef double z0 = max(z)    
    return z0 + log(sum(map(lambda x: exp(x - z0), z)))

#################################
#
#   Model Vector Class
#
################################
cdef inline check_add_key(d, k):
    if not (d.has_key(k)):
        d[k] = dict()

cdef inline check_path(d, ls):
    last_d = d
    for l in ls:
        check_add_key(last_d, l)
        last_d = last_d[l]
    return last_d 

cdef get_fail_path(d, ls, fail_value):
    last_d = d
    for l in ls:
        if last_d.has_key(l):
            last_d = last_d[l]
        else:
            return fail_value
    return last_d

##################
#
# ModelVector
#
class ModelVector:
    def __init__(self):
        ###################################################
        # unigrams: (cur_label, fv) -> real 
        #
        self.unigrams     = dict()
        ###################################################
        #
        # bigrams: (prev_label, cur_label, fv) -> real
        #
        self.bigrams      = dict()

        # We implement the scale factor separately as it makes
        # some proximal operators easier to implement.
        self.scale_factor = 1.0

    def get_unigram_feature(self, label, fv):
        return get_fail_path(self.unigrams, [label,fv], 0.0)

    def get_bigram_feature(self, l1, l2, fv):
        return get_fail_path(self.bigrams, [l1,l2,fv], 0.0)

    def get_bigram_feature_labels(self, l1, l2):
        return check_path(self.bigrams, [l1,l2])

    #########################
    # Add features
    #
    def add_unigram_feature(self, label, fv, double value=1.0):
        if not (self.unigrams.has_key(label)):
            self.unigrams[label] = dict()
        if not (self.unigrams[label].has_key(fv)):
            self.unigrams[label][fv] = 0.0
        self.unigrams[label][fv] += value

    def add_bigram_feature(self, l1, l2, fv, double value=1.0):
        leaf = check_path(self.bigrams, [l1,l2])
        if not (leaf.has_key(fv)):
            leaf[fv] = 0.0
        leaf[fv] += value
        
    def scale(self, double scale_factor):
        self.scale_factor = scale_factor
        return self
                        
    def dot(self, them):
        cdef double r
        r = 0.0
        for (l,d) in them.unigrams.iteritems():
            for (t,v) in d.iteritems():
                r += self.get_unigram_feature(l, t) * v

        for (l1,d1) in them.bigrams.iteritems():
            for (l2,d2) in d1.iteritems():
                for (fv,v) in d2.iteritems():
                        r += self.get_bigram_feature( l1,l2,fv) * v
        return r * self.scale_factor * them.scale_factor


    def add_sparse_features(self, them, scale=1.0):
        cdef double ratio = scale*them.scale_factor/self.scale_factor
        # Add in the unigrams
        for (l,d) in them.unigrams.iteritems():
            for (t,v) in d.iteritems():
                self.add_unigram_feature(l, t, value=v*ratio)
                
        for (l1,d1) in them.bigrams.iteritems():
            for (l2,d2) in d1.iteritems():
                for (fv,v) in d2.iteritems():
                    self.add_bigram_feature(l1, l2, fv, value=v*ratio)

    def ell2_project(self, B):
        norm = sqrt(self.dot(self))
        self.scale(sqrt(B)/norm)
        
    def add(self, them, scale=1.0):
        self.add_sparse_features(them, scale=scale)

    ## Print
    def print_model(self):
        header = "scale={0}".format(self.scale_factor)
        
        # Add in the unigrams
        features = []
        for (l,d) in self.unigrams.iteritems():
            for (t,v) in d.iteritems():
                features.append( "{0}/{1} {2}".format(l,t,self.get_unigram_feature(l, t)*self.scale_factor) )
                
        for (l1,d1) in self.bigrams.iteritems():
            for (l2,d2) in d1.iteritems():
                for (fv,d3) in d2.iteritems():
                        value = self.get_bigram_feature(l1, l2, fv)*self.scale_factor
                        if value > 1e-15:
                            features.append( "{0}/{1}/{2} {3}".format(l1,l2,fv,value) )
        return header + '\n'.join(features)        
    
    
### End of ModelVector Class
        
    

    
###############
#                                        
# Dot products
#
###############
def unlabeled_dot(mv, l_cur, cur_fv):
    return sum([mv.get_unigram_feature(l_cur, cf) for cf in cur_fv])

def unlabeled_bigram_dot(mv, last_label, cur_label, fvs):
    cdef double r = 0.0
    d = mv.get_bigram_feature_labels(last_label, cur_label)
    for fv in fvs:
        if d.has_key(fv):
            r += d[fv]
    return r

# We use reals instead of Booleans because we plan to whiten.
# type feature_value = string * real
# type LabeledItem   = (string * feature_value * label)
# Labeled Document is a LabeledItem list
def build_doc_vector(d):
    dv = ModelVector()
    last_label = None
    last_fvs   = None
    cdef int nfeatures = 0
    label_set = set()
    for (text, unigram_templates, bigram_templates, label) in d:
        # Add in the unigrams
        for fv in unigram_templates:
            dv.add_unigram_feature(label, fv)
        # Add in the bigrams
        for fv in bigram_templates:
            if last_label is not None:
                dv.add_bigram_feature(last_label, label, fv)                    

        last_label = label
        label_set.add(label)
    return (dv,label_set)                           

#############################################
#
# Compute the most likely path and its weight
#
#############################################
def compute_most_likely_path(m, d):
    mv        = m.model_vector
    label_set = m.label_set
    cdef int i
    
    w = [dict() for i in xrange(len(d))]
    (s,unigrams, bigrams,l0) = d[0]

    # Handle_initial condition
    for li in label_set:
        val      = (unlabeled_dot(mv, li,unigrams), None)
        w[0][li] = val
    
    # Viterbi
    for i in range(1,len(d)):
        (last_s,last_unis, last_bigrams,last_label) = d[i-1]
        (cur_s, cur_unis , cur_bigrams ,cur_label ) = d[i]
        for l_cur in label_set:
            for l_prev in label_set:
                score_prev_to_cur = w[i-1][l_prev][0] + unlabeled_bigram_dot(mv, l_prev, l_cur, cur_bigrams) + unlabeled_dot(mv, l_cur,cur_unis)
                if not w[i].has_key(l_cur):
                    w[i][l_cur] = (score_prev_to_cur,l_prev)
                else:
                    if w[i][l_cur][0] < score_prev_to_cur:
                        w[i][l_cur] = (score_prev_to_cur,l_prev)
                        
    # Viterbi Path
    path = ["a" for i in xrange(len(d))]
    last_index = "a"
    w0    = None # w0 will contain the final weight
    
    # find the maximum weight path
    for (last_label,(ws,last_ind)) in w[len(d) - 1].iteritems():
        if (w0 is None) or (w0 < ws):
            path[len(d)-1] = last_label
            last_index     = last_ind
            w0             = ws

    
    # now walk back through the linked list to find the full path
    # last_index contains the next place we are to visit.
    for i in range(len(d)-2,-1,-1):
        # add it to the path
        path[i] = last_label
        # if we have not yet reached the start,
        # look backward to find the next part of the path.
        if i > 0:
            (ws,last_label)=w[i][last_label]
    # return the path and its weight
    return (w0,path)

# Compute the weight of the labeled document d
def compute_weight_of_labeled(m, d):
    mv         = m.model_vector
    last_label = None
    cdef double weight     = 0.0
    
    for (s, current_unis, current_bigrams, current_label) in d:        
        weight += unlabeled_dot(mv, current_label, current_unis)
        if last_label is not None:
            weight += unlabeled_bigram_dot(mv, last_label, current_label, current_bigrams)
        last_label = current_label
    return weight

##################################
#
# compute the total weight (Z)
#
##################################
def compute_normalization(m, d):
    mv        = m.model_vector
    label_set = m.label_set
    cdef int i 

    w = [dict() for i in xrange(len(d))]
    (s,unis,bis,l0) = d[0]
    
    # Handle_initial condition
    for li in label_set:
        w[0][li] = unlabeled_dot(mv, li, unis)
    
    # Sum Product
    for i in range(1,len(d)):
        (last_s,last_unis, last_bis,last_label) = d[i-1]
        (cur_s ,cur_unis , cur_bis, cur_label)  = d[i]
        for l_cur in label_set:
            w[i][l_cur]      = exp_sum([ w[i-1][l_prev] + unlabeled_bigram_dot(mv, l_prev, l_cur, cur_bis) + unlabeled_dot(mv, l_cur, cur_unis) for l_prev in label_set ])
    
 
    return exp_sum([ v for (k,v) in w[len(d)-1].iteritems() ] )

#
# Compute a vector v(m,d) s.t.
# v[x]*Z = log(E[x]) according to m and d
#
def compute_expectation_vector(m, d):
    mv        = m.model_vector
    label_set = m.label_set
    if len(label_set) == 0:
        # we know that it is 0
        return ModelVector()

    cdef int i
    cdef double running_total
    
    forward_weights = [dict()for i in xrange(len(d)) ]
    
    edv = ModelVector()

    #####################
    # The Forward Pass
    #####################
    # Handle_initial condition
    (s,cur_unis,cur_bis,l0) = d[0]
    for li in label_set:
        forward_weights[0][li] = unlabeled_dot(mv, li, cur_unis)
  
    for i in range(1,len(d)):
        (last_s,last_unis,last_bis,last_label) = d[i-1]
        (cur_s, cur_unis, cur_bis,cur_label)   = d[i]
        for l_cur in label_set:
            q = [ forward_weights[i-1][l_prev] + unlabeled_bigram_dot(mv, l_prev, l_cur, cur_bis) + unlabeled_dot(mv, l_cur, cur_unis) for l_prev in label_set ]
            m = exp_sum(q)
            assert(m >= max(q))
            forward_weights[i][l_cur]      = m
    # We now have the normalization constant.
    Z = exp_sum([ v for (k,v) in forward_weights[len(d)-1].iteritems() ] )
                

    #######################
    # The Backward Phase
    #######################
    p = [dict() for i in xrange(len(d))]
    # print "Z={0} weights={1}".format(Z, forward_weights[len(d)-1])
    
    # Handle the final
    for li in label_set:
        p[len(d) - 1][li] = exp( forward_weights[len(d)-1][li] - Z )

    # Compute the probabilities for the remaining phases.
    # Terminology is that we have computed the next phase
    # and we want to compute the current phase (denoted by i)
    for i in xrange(len(d)-2,-1,-1):
        next_unis = d[i+1][1]
        next_bis  = d[i+1][2]
        next_weights = forward_weights[i+1]
        
        cur_unis  = d[i][1]
        cur_bis   = d[i][2]
        
        cur_weights = forward_weights[i]
        # initializae all the probabilities for this current phase to 0.
        for l_cur in label_set:
            p[i][l_cur] = 0.0
    
        for l_next in label_set:
            # Get the probability of this label
            marginal_next_label = p[i+1][l_next]
            # Get the weight of the next label
            log_weight_next        = next_weights[l_next]
            
            # Unigram weight of the next.
            log_l_uninext          = unlabeled_dot(mv, l_next, next_unis)
            for l_cur in label_set:
                # get ahold of this portion of the vector
                # This is just to avoid reaccessing this dictionary repeatedly.
                z = mv.get_bigram_feature_labels(l_cur, l_next)
        
                # Now compute the contribution
                # of the bigrams on the edge (cur, l_cur) -> (next, l_next)
                # check all the possible bigram contributors
                # Three components:
                # 1. the unigram weight on the next node
                # 2. the total weight on the current node
                # 3. the weight on the edge between them (the bigrams)
                paths_to_next    = log_l_uninext  + cur_weights[l_cur] + unlabeled_bigram_dot(mv, l_cur, l_next, next_bis)
                
                # assert(paths_to_next <= log_weight_next)
                # Now we compute the conditional probability of transitioning from
                # l_next to l_cur given that we are at l_next
                conditional_prob = exp(min(paths_to_next - log_weight_next,0))
                # ... and the marginal probability of taking this path
                path_prob        = conditional_prob * marginal_next_label
                # print "next_marg={0} conditional={1} path_prob={2}".format(marginal_next_label, conditional_prob, path_prob)
                # sum up this contribution.
                p[i][l_cur]     += path_prob
                
                # Since, we're here. We add up the expectations of all the bigram features
                # essentially adding the probability to each of their counts in edv.
                for (fv,v) in z.iteritems():
                    edv.add_bigram_feature(l_prev, l_cur, fv, value=v*path_prob)

        # Here, we know the marginal probability of each node at l_cur. 
        # finally add in the expectations for the local unigrams
        for l_cur in label_set:
            p_i_lcur = min(p[i][l_cur],1.0) 
            if p_i_lcur > 0.0:
                for fv in cur_unis:
                    edv.add_unigram_feature(l_cur, fv, value=p_i_lcur)
        
    return edv


#####
# These functions are external.
# The units we pass in and out are SimpleCRFModel and documents.
# Within this module, we know that (1) a SimpleCRFModel is a model and a vector.
#                                  (2) documents are represented as model vectors
#
# Build the Model
def build_model(document_gen,label_set,stepsize):
    mv = ModelVector()
    cdef int n  = 0
    label_set = set()
    for d in document_gen:
        (dv,d_label_set) = build_doc_vector(d)
        mv.add_sparse_features( dv )
        label_set.update(d_label_set)
        n += 1
    if n > 0:
        mv.scale(1.0/n)
    return crf_model.SimpleCRFModel(model_vector=mv, label_set=label_set, stepsize=stepsize, nSteps=[0])



#########################
# Compute Gradient
# The gradient is dv - edv - 1/sigma*m
# sigma=None signifies no regularization
#########################

def compute_gradient_step(m, d, sigma, debug_gradient=False):
    edv               = compute_expectation_vector(m,d)
    (dv,label_set)    = build_doc_vector(d)
    dv.add(edv,scale=-1.0)
    m.label_set.update(label_set)

    if sigma is not None:
        dv.add(m.model_vector, scale=-1.0/sigma)
    
    # debugging information
    if debug_gradient:
        variance = sqrt(dv.dot(dv))
        print "GS: variance={0}".format(variance)
    return dv

# BROKEN
# Computes the step
def l2_prox(m, double s):
    m.model_vector.scale(1.0/(1.0-s))
    
def l1_prox(m, double s):
    m.model_vector.squash(s)

# TODO: Write in the Entropy loss.
def no_prox(m, double s):
    0

def take_gradient_step(m, d, regularize=None, mu=1e-1, sigma=None):        
    steps = m.nSteps[0]
    m.model_vector.add_sparse_features(compute_gradient_step(m,d, sigma).scale(m.stepsize/(1.0 + steps)))        
    m.nSteps[0] = steps + 1
    if regularize is not None:
        regularize(m, mu*stepsize)
    return m

def batch_gradient(m, ds, stepsize, regularize=None, sigma=None):
    grad = ModelVector()
    for d in ds:
        grad.add_sparse_features( compute_gradient_step(m,d,sigma) )
    grad.scale(stepsize)
    m.model_vector.add_sparse_features(grad)
    if regularize is not None:
        regularize(m, mu*stepsize)

##################
#
# Document Scoring
#
##################
def score_documents(m, ds, verbose=False):
    cdef int correct = 0
    cdef int total   = 0
    cdef double obj     = 0.0
    for d in ds:
        t0        = time.time()
        # compute the normalization constant Z
        z         = compute_normalization(m, d)
        norm_t    = time.time()

        # compute the weight that the model currently 
        # assigns to the path in the ground truth
        v         = compute_weight_of_labeled(m, d)
        v_t       = time.time()

        # compute the path that is predicted as
        # most likely according to the model
        (wp,path) = compute_most_likely_path(m, d)
        mlp_t     = time.time()

        # Compute the true path explicitly for comparison
        tpath     = [ll for (s,u,b,ll) in d]
        
        if verbose:
            time_str = "[times={0} {1} {2}]".format(norm_t - t0, v_t - norm_t, mlp_t - v_t)
            print "{0} {1} {2}\n\t[v={5} z={6} wp={8}]\n\{7}\n\t{3}\n\t{4}".format(z, v, 1.0 - exp(v - z), path, tpath, v,z, time_str, wp)
        correct += sum([ tpath[i] == path[i] for i in xrange(len(path)) ])
        total   += len(path)
        obj     += z - v
    return (correct, total, obj)
