import numpy, math, time, array
import src.utils.shared_storage as ss
cdef float clip(float x):
    if x > 1.0:   return 1.0
    elif x < 0.0: return 0.0
    else:       return x

cdef extern from "numpy/arrayobject.h":
    ctypedef int intp
    ctypedef extern class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef int nd
        cdef intp *dimensions
        cdef intp *strides
        cdef int flags

    
class GraphModel:
    def __init__(self, nNodes, S, T, array):
        nNodes = max([nNodes,max(S), max(T)])
        self.nodes = array
        for s in S: self.nodes[s] = 0.0
        for t in T: self.nodes[t] = 1.0
        self.terminals     = set(S).union(set(T))
        self.stepsize      = 5e-2
        self.steps         = 1

    def c_grad(self, int x, int y, double w, double stepsize):
        if self.nodes[x] == self.nodes[y]:
            return

    
        diff = math.copysign(stepsize*w, self.nodes[x] - self.nodes[y])

        if x not in self.terminals: self.nodes[x] = clip(self.nodes[x] - diff)
        if y not in self.terminals: self.nodes[y] = clip(self.nodes[y] + diff)

    def grad(self, d):
        self.steps += 1
        cdef double stepsize    = 0.1/(1+self.steps)
        self.c_grad(d[0], d[1],d[2], stepsize)

    def loss(self, d):
        (x,y,w) = d
        return w*abs(self.nodes[x] - self.nodes[y])
        
def stopping_condition(s, loss):
    if not (s.has_key('state')):
        s['state'] = 0
        s['last_loss'] = loss.value
        return False
    else:
        s['state'] += 1
        print "\tlast_loss={} loss_value={} diff={}".format(s['last_loss'], loss.value, abs(s['last_loss'] - loss.value))
        done = (s['state'] > 1000) or (abs(s['last_loss'] - loss.value) < 1e-3)
        done = done and (s['state'] > 2)
        s['last_loss'] = loss.value
        return done

def init():
    import src.globals.global_utils as gu
    conn = gu.connect_db()
    cur  = conn.cursor()

    cur.execute("SELECT max(eid) FROM dblife_entities")
    (nNodes,) = cur.fetchone()
    conn.close()
    
    S     = range(10)
    T     = range(30,50)
    array = numpy.array([0.5]*nNodes)
    return (GraphModel(nNodes+1, S, T, array),array)

def grad(m,data):
    m[0].grad(data)
    return m

def loss(m,data):
    return m[0].loss(data)

def loss_two(m,d):
    (x,y,w) = d
    return w*abs(m[1][x] - m[1][y])

def minibatch(model, batch_data,perm):
    m = model[0]
    cdef int i, pi, nItems, minBatchSize
    minBatchSize = 20
    nItems = len(batch_data)
    x_data = array.array('i', map(lambda x:x[0], batch_data))
    y_data = array.array('i', map(lambda x:x[1], batch_data))
    w_data = array.array('d', map(lambda x:x[2], batch_data))
    cdef stepsize = m.stepsize
    for epoch in xrange(minBatchSize):
        for i in xrange(nItems):
            pi = perm[i] 
            m.c_grad(x_data[pi], y_data[pi], w_data[pi], stepsize)
        stepsize *= 0.99
    m.stepsize = stepsize




cdef c_grad_inline(double* nodes, int x , int y , double w , double stepsize , int x_term , int y_term ):
    if nodes[x] == nodes[y]:
        return
            
    diff = numpy.copysign(stepsize*w, nodes[x] - nodes[y])

    if x_term == int(0): nodes[x] = clip(nodes[x] - diff)
    if y_term == int(0): nodes[y] = clip(nodes[y] + diff)

def minibatch_two(model, batch_data,perm):
    m = model[0]
    cdef int i, pi, nItems, minBatchSize
    minBatchSize = 100
    nItems = len(batch_data)
    x_data = array.array('i', map(lambda x:x[0], batch_data))
    x_term = array.array('i', [0]*nItems)
        
    y_data = array.array('i', map(lambda x:x[1], batch_data))
    y_term = array.array('i', [0]*nItems)

    for i in xrange(nItems):
        if x_data[i] in m.terminals: x_term[i] = 1
        if y_data[i] in m.terminals: y_term[i] = 1
    
    weight_data = array.array('d', map(lambda x:x[2], batch_data))
    cdef double* wd = <double*> ((<ndarray>(model[1])).data)
    cdef stepsize = m.stepsize*(0.95)**((ss.get_epoch()*minBatchSize/2))
    for epoch in xrange(minBatchSize):
        for i in xrange(nItems):
            pi = i
            c_grad_inline(wd, x_data[pi], y_data[pi], weight_data[pi],
                          stepsize,
                          x_term[pi], y_term[pi])
        stepsize *= 0.95
    m.stepsize = stepsize
