#include "postgres.h"
#include "fmgr.h"
#include "utils/memutils.h"
#include "utils/dynamic_loader.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include "funcapi.h"
#include "init_helper.h"
#include <math.h>
#include <assert.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


PG_FUNCTION_INFO_V1(factor_partition);

#ifndef min
#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif

#ifndef max
#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

const char* SEED_STR="seed";
const char* ROW_PERM_STR="row_perm";
const char* COL_PERM_STR="col_perm";
const char* ROW_STR="row";
const char* COL_STR="col";
static void* gd = NULL;

void shuffle_array(int *arr, int dim) {
  int i;
  for(i = 0; i < dim; i ++) {
    int rand_ind  = rand() % dim;

    // swap the values
    int temp      = arr[rand_ind];
    arr[rand_ind] = arr[i];
    arr[i]        = temp;
  }
}

int*
alloc_int(int *p, int v) {
  int * ret = p;
  if(ret == NULL) { ret = malloc(sizeof(int)); }
  ret[0] = v;
  return ret;
}

int*
generate_permutation(int *r, int nSize) {
  int i;
  if(r != NULL) { free(r); } 
  int *ret = malloc(sizeof(int) * nSize);
  for(i = nSize - 1; i >= 0; i--) ret[i] = i;
  return ret;
}

Datum
factor_partition(PG_FUNCTION_ARGS)
{
  int nRows         = PG_GETARG_INT32(0);
  int nCols         = PG_GETARG_INT32(1);
  int seed          = PG_GETARG_INT32(2);
  int nSplits2      = PG_GETARG_INT32(3);
  HeapTupleHeader t = PG_GETARG_HEAPTUPLEHEADER(4);

  //elog(INFO, "Enter: %d", fcinfo->flinfo);

  //gd = !MemoryContextContains(TopTransactionContext, (void*) gd) ? NULL : gd;
  gd = check_init(gd);
  //elog(INFO, "gd=%d", (unsigned int) gd);
  int *seed_val, *_nRows, *_nCols, *row_perm, *col_perm;

  seed_val = get_key(gd, SEED_STR);
  _nRows    = get_key(gd, ROW_STR);
  _nCols    = get_key(gd, COL_STR);

  if(seed_val == NULL || *seed_val != seed || _nRows == NULL || _nCols == NULL
     || _nRows[0] != nRows || _nCols[0] != nCols) {
    //elog(INFO, "Rebuild");
    seed_val = alloc_int(seed_val, seed);
    _nRows   = alloc_int(_nRows,   nRows);
    _nCols   = alloc_int(_nCols,   nCols);

    row_perm = generate_permutation(get_key(gd, ROW_PERM_STR), nRows);
    col_perm = generate_permutation(get_key(gd, COL_PERM_STR), nCols);
    
    srand(seed);
    shuffle_array(row_perm, nRows);
    shuffle_array(col_perm, nCols);

    set_key(gd, SEED_STR, seed_val);
    set_key(gd, ROW_STR, _nRows);
    set_key(gd, COL_STR, _nCols);
    set_key(gd, ROW_PERM_STR, row_perm);
    set_key(gd, COL_PERM_STR, col_perm);
  }
 
  //elog(INFO, "\t seed_deref=%d seed=%d splits=%d", *seed_val, seed, nSplits2);
 
  row_perm = (int *) get_key(gd, ROW_PERM_STR);
  col_perm = (int *) get_key(gd, COL_PERM_STR);
  int nSplits = (int) sqrt(nSplits2);
  if(nSplits * nSplits != nSplits2) {
    elog(ERROR, "nSplits must be a perfect square!");
  }

  bool isnull;
  int r0 = DatumGetInt32(GetAttributeByNum(t, 1, &isnull)) - 1;   assert(!isnull);
  int c0 = DatumGetInt32(GetAttributeByNum(t, 2, &isnull)) - 1;   assert(!isnull);

  //elog(INFO, "\t r0=%d (%d) c0=%d (%d)", r0,nRows,c0,nCols);
  //elog(INFO, "rs=%d cs=%d", row_perm, col_perm);

  if (r0 >= nRows || c0 >= nCols || r0 < 0 || c0 < 0)
    elog(ERROR, "r0: %d, c0: %d", r0, c0);

  int r = row_perm[r0];
  int c = col_perm[c0];

  //elog(INFO, "\t r=%d c=%d", r,c);


  int r_val1 = ((int)((double) r / ((double) nRows / (double) nSplits)));
  int c_val1 = ((int)((double) c / ((double) nCols / (double) nSplits)));

  int maxSplitIndex = nSplits - 1;
  int row_bucket    = min(r_val1, maxSplitIndex);
  int col_bucket    = min(c_val1, maxSplitIndex);

  int b = max(min((row_bucket * nSplits + col_bucket), nSplits2 - 1), 0);
  //elog(INFO, "Exit=%d\n",b);

  PG_RETURN_INT32(b);
}
