"""
This module contains non incremental (batch or full gradient) functions.
"""

import src.vsql_core.utils.victor_utils as victor_utils, math

########################
## Full Gradient Code
# 
########################
#################
# Ell2 gradient
###############
def ell2_full_gradient(w, cur_grad, x, y):
   """
   Calculates ell2 gradient.
   
   @type w: vector
   @param w: model vector
   @type cur_grad: vector
   @param cur_grad: current gradient
   @type x: vector
   @param x: feature vector of the example
   @type y: number
   @param y: label of the example
   """
   wx  = victor_utils.dot(w, x)
   err = (wx - y)
   victor_utils.scale_and_add_i(cur_grad, x, err)
   
# ell2 gradient step
def ell2_gradient_step(w, grad, stepsize, mu, B):
   """
   Computes w as a step
   @type w: vector
   @param w: model vector
   @type grad: vector
   @param grad: gradient
   @type stepsize: double
   @param stepsize: step size
   @type mu: double
   @param mu: mu
   @type B: double
   @param B: B
   """
   # compute w as a step
   victor_utils.scale_and_add_i(w, grad, -stepsize)
   victor_utils.l2_project(w, B)

################
# SVM + L2 Prox
##############
def svm_full_gradient(w, cur_grad, x,y):
   """
   Calculates svm gradient.
   
   @type w: vector
   @param w: model vector
   @type cur_grad: vector
   @param cur_grad: current gradient
   @type x: vector
   @param x: feature vector of the example
   @type y: number
   @param y: label of the example
   """
   wx  = victor_utils.dot(w, x)
   if(y*wx < 1):
      victor_utils.scale_and_add_i(cur_grad, x, y)

def ell2_prox_term(w, cur_grad, stepsize, mu, B):
   """
   Computes w as a step
   @type w: vector
   @param w: model vector
   @type cur_grad: vector
   @param cur_grad: gradient
   @type stepsize: double
   @param stepsize: step size
   @type mu: double
   @param mu: mu
   @type B: double
   @param B: B
   """
   victor_utils.scale_and_add_i(w, cur_grad, -stepsize)
   victor_utils.scale(w, 1.0/(1.0 + stepsize * mu))


################
# Logit helpers
##############
def logit_l1_sparse(w, cur_grad, x_indexes, x_vectors, y):
   """
   Calculates logit l1 sparse
   @type w: vector
   @param w: model vector
   @type cur_grad: vector
   @param cur_grad: current gradient
   @type x_indexes: vector
   @param x_indexes: indexes vector of the example
   @type x_vectors: vector
   @param x_vectors: values of the feature vector of the example
   @type y: number
   @param y: label of the example
   """
   wx      = victor_utils.dot_dss(w,x_indexes,x_vectors)
   sig     = victor_utils.sigma(-wx*y)
   victor_utils.scale_and_add_dss(cur_grad, x_indexes, x_vectors, -y*sig)

def l1_shrink_prox(w, cur_grad, stepsize, mu, B):
   """
   Computes l1 shrink prox
   @type w: vector
   @param w: model vector
   @type cur_grad: vector
   @param cur_grad: gradient
   @type stepsize: double
   @param stepsize: step size
   @type mu: double
   @param mu: mu
   @type B: double
   @param B: B
   """
   victor_utils.scale_and_add_di(w,cur_grad, -stepsize)
   victor_utils.l1_shrink_mask(w, mu*step)

def logit_loss(w, x_indexes, x_values, y):   
   """
   Returns logit loss
   @type w: vector
   @param w: model vector
   @type x_indexes: vector
   @param x_indexes: indexes vector of the example
   @type x_values: vector
   @param x_values: values of the feature vector of the example
   @rtype: double
   @return: logit loss of the given model and the example
   """
   wx      = victor_utils.dot_dss(w,x_indexes,x_values)
   return  math.log(1 + math.exp(-y*wx))
