set -o nounset #exit the script if it tries to use an uninitialized variable
set -o errexit #force an exit if any statement returns a non-true value.
set -o pipefail #fail if any item in a pipeline fails.

###########################
#Header Variables to update
###########################
#_tar variable: the name of the tar file for the package
#_folder variable: name of the untarred package folder and the installation folder in case of Python and PostgreSQL
python_tar="Python-2.7.tgz.tar"
python_folder="python-2.7.1"
postgres_tar="postgresql-9.0.4.tar.gz"
postgres_folder="postgresql-9.0.4"
cython_tar="Cython-0.14.1.tar.gz"
cython_folder="Cython-0.14.1"
antlr_tar="antlr-3.3.tar.gz"
antlr_folder="antlr-3.3"
psycopg2_tar="psycopg2-2.4.tar.gz"
psycopg2_folder="psycopg2-2.4"
numpy_tar="numpy-1.5.1.tar.gz"
numpy_folder="numpy-1.5.1"
#setuptools compatible with your ANTLR version. If you are not sure about the version, make your best guess, 
#if it is not the right file, installation will fail and let you know the version that it expects.
setuptools="setuptools-0.6c5-py2.7.egg"
########################
#END of Header Variables
########################

#absolute path to this script.
exec_path=`readlink -f $0`
#absolute path to the folder that this script is in.
exec_dir_name=`dirname $exec_path`
#Victor base folder. Absolute path to the folder where the setup folder is in.
victor_base_dir="$exec_dir_name/.."

# <base_dir>: Your installation folder path. Folder where you want your packages to be installed in.
if [ $# -ne 1 ]
then
	echo "Usage: $0 <base_dir>"
	exit
fi 

#create the installation folder if it does not exist.
if [ ! -d "$1" ]; then
	echo "Creating installation directory: $1"
	mkdir -p $1
fi

base=$1

path_file="$victor_base_dir/victor.paths.src"
#remove path file if exists
rm -rf $path_file
#go to the setup folder
cd $victor_base_dir/setup/

###############
#install Python
###############
echo "Installing Python."
rm -rf $python_folder
mkdir $python_folder
tar zxvf $python_tar -C $python_folder --strip-components=1
cd $python_folder
python_path=$base/$python_folder
rm -rf $python_path
mkdir $python_path
./configure --prefix=$python_path --enable-shared
make || { echo "Python make step failed."; exit; }
make install || { echo "Python installation step failed."; exit; }
echo "export PATH=$python_path/bin:\$PATH" >> $path_file
echo "export LD_LIBRARY_PATH=$python_path/lib:\$LD_LIBRARY_PATH" >> $path_file
echo "Python installed successfully."
#back to the setup folder 
cd ../

###################
#install Postgresql
###################
echo "Installing PostgreSQL."
rm -rf $postgres_folder
mkdir $postgres_folder
tar zxvf $postgres_tar -C $postgres_folder --strip-components=1 
cd $postgres_folder
postgres_path=$base/$postgres_folder
rm -rf $postgres_path
mkdir $postgres_path
./configure --prefix=$postgres_path --with-python
make || { echo "PostgreSQL make step failed."; exit; }
make install || { echo "PostgreSQL installation step failed."; exit; }

echo "export PATH=$postgres_path/src/bin:$victor_base_dir/setup/$postgres_folder/src/bin/pg_config:\$PATH" >> $path_file
echo "export LD_LIBRARY_PATH=$postgres_path/lib:\$LD_LIBRARY_PATH" >> $path_file
#back to the setup folder
cd ../
set +u #uninitialized variables are ok for export.
source $path_file
set -u
echo "PostgreSQL installed successfully."

###############
#install Cython
###############
echo "Installing Cython."
rm -rf $cython_folder
mkdir $cython_folder
tar zxvf $cython_tar -C $cython_folder --strip-components=1
cd $cython_folder
python setup.py install || { echo "Cython installation failed."; exit; }  
echo "Cython installed successfully."
#back to the setup folder
cd ../

#################
#install psycopg2
#################
echo "Installing psycopg2."
rm -rf $psycopg2_folder
mkdir $psycopg2_folder
tar zxvf $psycopg2_tar -C $psycopg2_folder --strip-components=1
cd $psycopg2_folder
python setup.py install || { echo "psycopg2 installation failed."; exit; }
echo "psycopg2 installed successfully."
#back to the setup folder
cd ../

##############
#install antlr
##############
echo "Installing ANTLR."
rm -rf $antlr_folder
mkdir $antlr_folder
tar zxvf $antlr_tar -C $antlr_folder --strip-components=1
cp $setuptools $antlr_folder/runtime/Python/
cd $antlr_folder/runtime/Python/
python setup.py install || { echo "ANTLR installation failed."; exit; }
echo "ANTLR installed successfully."
#back to the setup folder
cd ../../../

##############
#install numpy
##############
echo "Installing Numpy."
rm -rf $numpy_folder
mkdir $numpy_folder
tar zxvf $numpy_tar -C $numpy_folder --strip-components=1 
cd $numpy_folder
python setup.py install || { echo "Numpy installation failed."; exit; }
echo "Numpy installed successfully."
#back to the Victor base folder
cd ../../
make clean
make || { echo "Victor make step failed."; exit; }
echo "export PYTHONPATH=$victor_base_dir:\$PYTHONPATH" >> $path_file
set +u
source $path_file
set -u
echo "Success: Victor package installation complete."
