"""
This module implements binary read/write methods
for atomic data types(non-array, 1d and 2d array types)
"""

import struct
import array
from victor_buffer cimport *
from data_types import *
import collections

TupleWithPartID = collections.namedtuple('TupleWithPartitionID', 'partition_id num_field tuple_data')

# global struct parsers
struct_char   = struct.Struct(oid_dict[charoid][data_struct_type_field])
struct_bool   = struct.Struct(oid_dict[booloid][data_struct_type_field])
struct_int2   = struct.Struct(oid_dict[int2oid][data_struct_type_field])
struct_int4   = struct.Struct(oid_dict[int4oid][data_struct_type_field])
struct_int8   = struct.Struct(oid_dict[int8oid][data_struct_type_field])
struct_float4 = struct.Struct(oid_dict[float4oid][data_struct_type_field])
struct_float8 = struct.Struct(oid_dict[float8oid][data_struct_type_field])

cdef class Atomic_Data_Processor:
    cdef int debug

    def __init__(self, debug=False):
        self.debug = debug

    #################################################
    # READ METHODS
    #################################################

    #NB: Is it inefficient to call oid_dict[charoid][data_size_field] instead of materalizing them?


    # TODO: Is this inefficient to be called inside read_type method?
    """
    Generic read non-array type method
    it takes unpack function and size of the type to advance
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @type fnc: function
    @param fnc: unpack function for the non-array type
    @type size_of_type: number
    @param size_of_type: size of the non-array type
    @rtype: data
    @return: the non-array data read from the buffer
    """
    cpdef read_non_array_type(self, VictorBuffer buf, fnc, size_of_type):
        # TODO: retrieve size and assert its size of type?
        buf.advance_ci(size_int) #advance length
        (v,) = fnc(buf.b, buf.current_index)
        buf.advance_ci(size_of_type)
        return v

    """
    Reads char type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: char
    @return: char read from the buffer
    """
    cpdef read_char(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_char.unpack_from, oid_dict[charoid][data_size_field])

    """
    Reads bool type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: bool
    @return: bool read from the buffer
    """
    cpdef read_bool(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_bool.unpack_from, oid_dict[booloid][data_size_field])

    """
    Reads int2 type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: int2
    @return: int2 read from the buffer
    """
    cpdef read_int2(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_int2.unpack_from, oid_dict[int2oid][data_size_field])

    """
    Reads int4 type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: int4
    @return: int4 read from the buffer
    """
    cpdef read_int4(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_int4.unpack_from, oid_dict[int4oid][data_size_field])

    """
    Reads only int2 type without reading its length
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: int2
    @return: int2 read from the buffer
    """
    cpdef read_int2_without_length(self, VictorBuffer buf):
        (v,) = struct_int2.unpack_from(buf.b, buf.current_index)
        buf.advance_ci(size_int2)
        return v

    """
    Reads only int4 type without reading its length
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: int4
    @return: int4 read from the buffer
    """
    cpdef read_int4_without_length(self, VictorBuffer buf):
        (v,) = struct_int4.unpack_from(buf.b, buf.current_index)
        buf.advance_ci(size_int)
        return v

    """
    Reads int8 type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: int8
    @return: int8 read from the buffer
    """
    cpdef read_int8(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_int8.unpack_from, oid_dict[int8oid][data_size_field])

    """
    Reads float4 type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: float4
    @return: float4 read from the buffer
    """
    cpdef read_float4(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_float4.unpack_from, oid_dict[float4oid][data_size_field])

    """
    Reads float8 type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: float8
    @return: float8 read from the buffer
    """
    cpdef read_float8(self, VictorBuffer buf):
        return self.read_non_array_type(buf, struct_float8.unpack_from, oid_dict[float8oid][data_size_field])


    """
    Generic read array type method
    it takes buffer, element type of array and read function to read the element type
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @type typ: oid
    @param typ: element type of the array
    @type read_fnc: function
    @param read_fnc: the read function to read non-array data type
    @rtype: data
    @return: the array data read from the buffer
    """
    cpdef read_array(self, VictorBuffer buf, typ, read_fnc):
        buf.advance_ci(size_int) #advance length
        cdef int array_size
        ndim       = 1
        buf.advance_ci(12) #skip three fields: ndim, has_null, elemtype
        
        array_size = self.read_int4_without_length(buf)
        lbound    = self.read_int4_without_length(buf)

        ret = array.array(typ, xrange(array_size))

        # read the actual data
        for i in xrange(array_size):
            ret[i] = read_fnc(buf)

        return ret

    # read char array
    #NB: When creating array by array.array function,
    #array.array('c', xrange(size)) does not work, since xrange(size) returns int list
    #therefore, we need to do chr(i) for i in xrange(size)
    #since this is a special case for char, it is coded in here (not calling the generic function)
    """
    Reads char array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the char array data read from the buffer
    """
    cpdef read_char_array(self, VictorBuffer buf):
        buf.advance_ci(size_int) #advance length
        cdef int array_size
        ndim       = 1
        buf.advance_ci(12) #skip three fields: ndim, has_null, elemtype
        
        array_size = self.read_int4_without_length(buf)
        lbound    = self.read_int4_without_length(buf)

        ret = array.array(oid_dict[charoid][data_python_type_field], [chr(i) for i in xrange(array_size)])

        # read the actual data
        for i in xrange(array_size):
            ret[i] = self.read_char(buf)

        return ret

    """
    Reads bool array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the bool array data read from the buffer
    """
    cpdef read_bool_array(self, VictorBuffer buf):
        return self.read_array(buf, oid_dict[booloid][data_python_type_field], self.read_bool)

    """
    Reads int2 array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the int2 array data read from the buffer
    """
    cpdef read_int2_array(self, VictorBuffer buf):
        return self.read_array(buf, oid_dict[int2oid][data_python_type_field], self.read_int2)

    """
    Reads int4 array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the int4 array data read from the buffer
    """
    cpdef read_int4_array(self, VictorBuffer buf):
        return self.read_array(buf, oid_dict[int4oid][data_python_type_field], self.read_int4)

    # read int8 array
    # NB: Python does not allow to create int8 arrays:
    # See http://docs.python.org/library/array.html
    # Therefore, ret_array should be created as [] and used append function
    # Since this is slower than array, read_generic_array function stays same, but
    # only read_int8_array function uses this method
    """
    Reads int8 array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the int8 array data read from the buffer
    """
    cpdef read_int8_array(self, VictorBuffer buf):
        buf.advance_ci(size_int) #advance length
        cdef int array_size
        ndim       = 1
        buf.advance_ci(12) #skip three fields: ndim, has_null, elemtype
        
        array_size = self.read_int4_without_length(buf)
        lbound    = self.read_int4_without_length(buf)

        ret = []

        # read the actual data
        for i in xrange(array_size):
            ret.append(self.read_int8(buf))

        return ret

    """
    Reads float4 array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the float4 array data read from the buffer
    """
    cpdef read_float4_array(self, VictorBuffer buf):
        return self.read_array(buf, oid_dict[float4oid][data_python_type_field], self.read_float4)

    """
    Reads float8 array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: array
    @return: the float8 array data read from the buffer
    """
    cpdef read_float8_array(self, VictorBuffer buf):
        return self.read_array(buf, oid_dict[float8oid][data_python_type_field], self.read_float8)

    """
    Reads float8 2d array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @rtype: 2d array
    @return: the float8 2d array data read from the buffer
    """
    cpdef read_float8_2d_array(self, VictorBuffer buf):
        cpdef int n0, n1, ndim

        buf.advance_ci(size_int) #advance length
        ndim = self.read_int4_without_length(buf)
        assert(ndim == 2)
        buf.advance_ci(8) #skip two fields: has_null, elemtype
        
        dims = []
        nelemnts = 1
        # evaluation order matters, so comprehension is cryptic
        for i in xrange(ndim):
            d = self.read_int4_without_length(buf)
            lbound = self.read_int4_without_length(buf)
            dims.append( (d, lbound) )

        n0 = dims[0][0]
        n1 = dims[1][0]

        ret = [[0.0] * n1 for i in xrange(n0)]

        for i in xrange(n0):
            for j in xrange(n1):
                ret[i][j] = self.read_float8(buf)

        return ret


    cpdef read_partition_tuple(self, VictorBuffer buf):
        nFields = self.read_int2_without_length(buf)
        if nFields < 1:
            return None
        part_id = self.read_int4(buf)
        nTupleFields = nFields - 1
        tuple_length = 0
        start_index = buf.current_index
        for i in range(nTupleFields):
            length_of_field = self.read_int4_without_length(buf)
            buf.advance_ci(length_of_field)
            tuple_length += size_int + length_of_field

        return TupleWithPartID(partition_id = part_id, 
                               num_field    = nTupleFields, 
                               tuple_data   = buf.b[start_index:start_index + tuple_length])

    #################################################
    # WRITE METHODS
    #################################################
    """
    Generic write non-array type method
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: data
    @param val: the data we write
    @type fnc: function
    @param fnc: pack function
    @type size_of_type: number
    @param size_of_type: size of the data type that we write
    """
    cpdef write_non_array_type(self, buf, val, fnc, size_of_type):
        buf.write(struct_int4.pack(size_of_type))
        buf.write(fnc(val))
        return size_of_type + 4

    """
    Writes char type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: char
    @param val: the data we write
    """
    cpdef write_char(self, buf, val):
        return self.write_non_array_type(buf, val, struct_char.pack, oid_dict[charoid][data_size_field])

    """
    Writes bool type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: bool
    @param val: the data we write
    """
    cpdef write_bool(self, buf, val):
        return self.write_non_array_type(buf, val, struct_bool.pack, oid_dict[booloid][data_size_field])

    """
    Writes int2 type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: int2
    @param val: the data we write
    """
    cpdef write_int2(self, buf, val):
        return self.write_non_array_type(buf, val, struct_int2.pack, oid_dict[int2oid][data_size_field])

    """
    Writes int4 type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: int4
    @param val: the data we write
    """
    cpdef write_int4(self, buf, val):
        return self.write_non_array_type(buf, val, struct_int4.pack, oid_dict[int4oid][data_size_field])

    """
    Writes int8 type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: int8
    @param val: the data we write
    """
    cpdef write_int8(self, buf, val):
        return self.write_non_array_type(buf, val, struct_int8.pack, oid_dict[int8oid][data_size_field])

    """
    Writes float4 type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: float4
    @param val: the data we write
    """
    cpdef write_float4(self, buf, val):
        return self.write_non_array_type(buf, val, struct_float4.pack, oid_dict[float4oid][data_size_field])

    """
    Writes float8 type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type val: float8
    @param val: the data we write
    """
    cpdef write_float8(self, buf, val):
        return self.write_non_array_type(buf, val, struct_float8.pack, oid_dict[float8oid][data_size_field])

    """
    Generic write array type method
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the array data we write
    @type elm_size: number
    @param elm_size: size of the element in the array that we write
    @type write_elem_fnc: function
    @param write_elem_fnc: the write function to write non-array data type
    """
    cpdef write_array(self, buf, arr, elm_size, elem_type, write_elem_fnc):
        arr_length = size_int * 5 + len(arr) * (size_int + elm_size)
        buf.write(struct_int4.pack(arr_length)) #Field size
        buf.write(struct_int4.pack(1)) #ndims
        buf.write(struct_int4.pack(0)) #has_null
        buf.write(struct_int4.pack(elem_type)) #elem_type

        buf.write(struct_int4.pack(len(arr))) #length
        buf.write(struct_int4.pack(1)) #lbound
        for val in arr:
            write_elem_fnc(buf, val)
        return arr_length + 4
        

    """
    Writes char array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the char array data we write
    """
    cpdef write_char_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[charoid][data_size_field], charoid, self.write_char)

    """
    Writes bool array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the bool array data we write
    """
    cpdef write_bool_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[booloid][data_size_field], booloid, self.write_bool)

    """
    Writes int2 array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the int2 array data we write
    """
    cpdef write_int2_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[int2oid][data_size_field], int2oid, self.write_int2)

    """
    Writes int4 array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the int4 array data we write
    """
    cpdef write_int4_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[int4oid][data_size_field], int4oid, self.write_int4)

    """
    Writes int8 array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the int8 array data we write
    """
    cpdef write_int8_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[int8oid][data_size_field], int8oid, self.write_int8)

    """
    Writes float4 array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the float4 array data we write
    """
    cpdef write_float4_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[float4oid][data_size_field], float4oid, self.write_float4)

    """
    Writes float8 array type to the buffer
    @type buf: buffer
    @param buf: the buffer we write to
    @type arr: array
    @param arr: the float8 array data we write
    """
    cpdef write_float8_array(self, buf, arr):
        return self.write_array(buf, arr, oid_dict[float8oid][data_size_field], float8oid, self.write_float8)

    """
    Writes float8 2d array type method
    @type buf: VictorBuffer
    @param buf: the buffer we read from
    @type arr: 2d array
    @param arr: the float8 2d array data that we write to the buffer
    """
    cpdef write_float8_2d_array(self, buf, arr):
        rows = len(arr)
        rank = len(arr[0])
        # array header is 3 ints
        # Then 4 ints to describe the bounds
        L_size = size_int * (3 + 4) + rows * rank * (size_int + size_float8)
        ndims = 2
        has_nulls = 0
        for x in [L_size, ndims, has_nulls, float8oid]:
            buf.write(struct_int4.pack(x))

        # Describe the array bounds
        for x in [rows, 1, rank, 1]:
            buf.write(struct_int4.pack(x))

        # write the elements
        for l in arr:
            for x in l:
                self.write_float8(buf, x)

        return L_size + 4

    cpdef write_partition_tuple(self, buf, tuple_data):
        buf.write(tuple_data)
    


    
