import profile
import sys
import psycopg2
import struct
import cStringIO, array, time

# OID TABLE from pg_type.h
# oid, monkiker, python type (including endian), byte size
oid_table = [
    (18, "CHAROID", 'c','c', 1),
    (16, "BOOLOID", 'b', 'b', 1),
    (20, "Int8OID", '>q', 'q', 8),
    (21, "Int2OID", '>h', 'h', 2),
    (23, "Int4OID", '>i', 'i', 4), 
    (700,"Float4OID", '>f', 'f', 4),
    (701, "Float8OID", '>d', 'd', 8),
    ]
oid_dict = dict()
for  (oid,moniker,struct_type,python_type,size) in oid_table:    
    oid_dict[oid] = (moniker,struct_type,python_type,size,struct.Struct(struct_type))

numeric_array_oids = [
    (1007, "INT4ARRAYOID"),
    (1021,"FLOAT4ARRAYOID"),
    (2211, "REGTYPEARRAYID"),
    (2277, "ANYARRAYOID")]

# global struct parsers
struct_int32  = struct.Struct('>i')
struct_int64  = struct.Struct('>q')
struct_int16  = struct.Struct('>h')
struct_double = struct.Struct('>d')

def get_type_length(oid):
    (moniker,struct_type,python_type,size,s) = oid_dict[oid]
    return size

# get the translation string for struct
def get_struct_python_type(elemtype):
    (moniker,struct_type,python_type,size,s) = oid_dict[elemtype]
    return struct_type

# get the translation string for struct
def get_struct_function(elemtype):
    (moniker,struct_type,python_type,size,s) = oid_dict[elemtype]
    return s


# get the stored type of an oid
def get_python_type(oid):
    (moniker,struct_type,python_type,size,s) = oid_dict[oid]
    return python_type


## From PSQL Webpage
# http://www.postgresql.org/docs/9.0/static/sql-copy.html
#
# To determine the appropriate binary format for the actual tuple data
# you should consult the PostgreSQL source, in particular the *send
# and *recv functions for each column's data type (typically these
# functions are found in the src/backend/utils/adt/ directory of the
# source distribution).


class BAIO:
    def __init__(self, b):
        self.b = b
        self.index = 0
        self.len   = len(b)
    def write(self, x):
        max_len = min(self.len, self.index + len(x))
        for i in xrange(self.index, max_len):
            self.b[i] = x[i]
    def contents(self):
        return self.b

psql_header = bytearray(b'PGCOPY\n') + bytearray(b'\377\r\n\0')

cdef class PSQLByteParser:
    cdef int max_len, current_index
    cdef int debug
    cdef bytes b
    
    def __init__(self, b, debug=False,checkHeader=True):
        self.b       = b
        self.debug   = debug
        self.max_len = len(b)
        self.current_index = 0
        if checkHeader:
            if len(b) <= 11 or not (b[0:11] == psql_header):
                print "lenb={1} b={0}".format(b,len(b))
                assert(False)
                
            self.current_index = 11 # skip the header
            flags_field   = self.read_int32()
            header_extension_length = self.read_int32()
            self.advance_ci(header_extension_length) # skip over the remaining 0 part of the header
            if debug:
                print "[Parsed Header] flags={0} header={1}".format(flags_field,header_extension_length)

        
        # self it points to tuples

    cpdef advance_ci(self, int i):
        self.current_index = self.current_index + i
        
    # Integers are in network byte order
    cpdef int read_int32(self):
        cdef int v
        (v,) = struct_int32.unpack_from(self.b, self.current_index)
        self.advance_ci(4)
        return v

    # Integers are in network byte order
    cpdef long read_int64(self):
        (v,) = struct_int64.unpack_from(self.b, self.current_index)
        self.advance_ci(8)
        return v


    cpdef short read_int16(self):
        (v,) = struct_int16.unpack_from(self.b, self.current_index)
        self.advance_ci(2)
        return v


    # Integers are in network byte order
    cpdef double read_double(self):
        cdef double v
        (v,) = struct_double.unpack_from(self.b, self.current_index)
        self.advance_ci(8)
        return v
    

    # return an element of oid type
    def read_psqltype(self,int oid):
        length  = get_type_length(oid)
        sf      = get_struct_function(oid)
        (v,)    = sf.unpack_from(self.b, self.current_index)
        self.advance_ci(length)        
        return v
        
    # arrayfuncs.c: 1523 9.0
    def read_array(self, max_length):
        ndim       = self.read_int32()
        has_null   = self.read_int32()
        elemtype   = self.read_int32() # NB: ASSUMING OID SIZE == 4        #
        
        dims       = []
        # evaluation order matters, so comprehension is cryptic.
        for i in xrange(ndim):
            d      = self.read_int32()
            lbound = self.read_int32()
            dims.append( (d,lbound) )
        
        # FOR NOW 1d array only
        assert(len(dims) == 1)
        array_size = dims[0][0]
        # HACK To test
        
        ret = array.array(get_python_type(elemtype), xrange(array_size))
                
        # Now read the actual data
        for i in xrange(array_size):
            d_length = self.read_int32()
            ret[i]   = self.read_psqltype(elemtype)
        return ret

    
    def read_binary_2d_double_array(self):
        if (self.current_index >= self.max_len):
            return None
        ndim       = self.read_int32()
        if (ndim != 2):
            return None
        has_null   = self.read_int32()
        elemtype   = self.read_int32() # NB: ASSUMING OID SIZE == 4        #

        dims       = []
        nelemnts   = 1
        # evaluation order matters, so comprehension is cryptic.
        for i in xrange(ndim):
            d      = self.read_int32()
            lbound = self.read_int32()
            dims.append( (d,lbound) )
        n0 = dims[0][0]
        n1 = dims[1][0]

        ret = [[0.0]*n1 for i in xrange(n0)]

        assert(ndim == 2)
        for i in xrange(n0):
            for j in xrange(n1):
                d_l = self.read_int32()
                ret[i][j] = self.read_double()
        return ret

    def read_all_2darrays(self):
        ret = []
        if self.debug:
            print "ci={0} max_i={1}".format(self.current_index, self.max_len)
        r = self.read_binary_2d_double_array()
          
        while(r):
            ret.append(r)
            r = self.read_binary_2d_double_array()
        return ret

    cpdef read_single_pair_2d(self):
        if self.debug:
            print "ci={0} max_i={1}".format(self.current_index, self.max_len)
        nFields = self.read_int16()
        if nFields <= 0:
            return None;

        f_length = self.read_int32()
        L     = self.read_binary_2d_double_array()

        f_length = self.read_int32()
        R      = self.read_binary_2d_double_array()
        return (L,R)

    # arrayfuncs.c: 1523 9.0
    cpdef read_double_array(self, max_length):
        cdef int ndim, has_null, elem_type, i, array_size
        ndim       = 1
        self.advance_ci(12)
        # ndim       = self.read_int32()
        # has_null   = self.read_int32()
        # elemtype   = self.read_int32() # NB: ASSUMING OID SIZE == 4        #
        
        # evaluation order matters, so comprehension is cryptic.

        array_size      = self.read_int32()
        lbound          = self.read_int32()
        
        ret = array.array('d', xrange(array_size))
                
        # Now read the actual data
        for i in xrange(array_size):
            self.advance_ci(4)
            ret[i]   = self.read_double()
        return ret

    cpdef read_int_array_pair(self):
        cdef int nFields, f_length, i, r_id
        cdef double rating
        nFields = self.read_int16()
        if nFields <= 0:
            return None;

        f_length = self.read_int32()
        r_id     = self.read_int32()

        f_length = self.read_int32()
        array    = self.read_double_array(f_length)
        return (r_id,array)
    
    cpdef read_all_ia_pairs(self):
        ret = []
        r = self.read_int_array_pair()
            
        while(r):
            ret.append(r)
            r = self.read_int_array_pair()
        return ret
    
    def read_bin_data(self,l):
        max_len = min(self.current_index + l, self.max_len)
        ci      = self.current_index
        self.advance_ci(l)
        return self.b[ci:max_len]
    
    cpdef read_triple(self):
        cdef int nFields, f_length, i, row, col
        cdef double rating
        nFields = self.read_int16()
        if nFields <= 0:
            return None
        assert(nFields >= 3)
        f_length = self.read_int32()        
        row      = self.read_int32()

        f_length = self.read_int32()
        col      = self.read_int32()

        f_length = self.read_int32()
        rating   = self.read_double()
        
        # skip over the rest
        for i in xrange(3,nFields):
            f_length = self.read_int32()
            self.advance_ci(f_length)
        return (row,col,rating)

    cpdef read_all_triples(self):
        ret = []
        r = self.read_triple()
            
        while(r):
            ret.append(r)
            r = self.read_triple()
        return ret
    
    # Just read a single array
    # We assume that is the first part coming back
    cpdef read_array_tuple(self):
        cdef int nFields, f_length, i
        nFields = self.read_int16()
        if nFields <= 0:
            return None
        
        f_length = self.read_int32()        
        r = self.read_double_array(f_length)
        # skip over the rest
        for i in xrange(1,nFields):
            f_length = self.read_int32()
            self.advance_ci(f_length)
        return r


    cpdef read_all_array_tuples(self):
        ret = []
        if self.debug:
            print "ci={0} max_i={1}".format(self.current_index, self.max_len)

        r = self.read_array_tuple()
        while(self.current_index < self.max_len and r):
            if r:
                ret.append(r)
            r = self.read_array_tuple()
        return ret

        
    def read_tuple(self):
        nFields = self.read_int16()
        if self.debug:
            print "nFields={0}".format(nFields)
            
        ret     = []
        for i in xrange(int(nFields)):
            f_length = self.read_int32()
            if self.debug:
                print "Field Length={0}".format(f_length)
            if f_length > 0:
                #b        = self.read_bin_data(f_length)
                b         = self.read_array(f_length)
                ret.append([b])
            else:
                ret.append([])
        return ret
    

def test():
    t = time.clock()
    conn = psycopg2.connect('dbname=chrisre')
    cur = conn.cursor()
    b = cStringIO.StringIO()
    cur.copy_expert("COPY bens_forest (vec) TO STDOUT WITH BINARY", b)
    v = time.clock() - t
    print "Time to execute copy query = {0}".format(v)

    t = time.clock()
    counter = 0
    bv   = b.getvalue()
    s_bv = len(bv)
    pb   = PSQLByteParser(bv)
    v0   = time.clock() - t
    for v in pb.read_all_array_tuples():
        counter = counter + 1    
    v = time.clock() - t
    print "Counter={0} timer={1}s tup/s={2} MB/s={4} build={3}".format(counter,v, counter/v, v0, s_bv/1e6/v)

def test2():
    t = time.clock()
    conn = psycopg2.connect('dbname=chrisre')
    cur = conn.cursor()
    b = cStringIO.StringIO()
    cur.copy_expert("COPY lr_data(row,col,rating) TO STDOUT WITH BINARY", b)
    v = time.clock() - t
    print "Time to execute copy query = {0}".format(v)

    t = time.clock()
    counter = 0
    pb = PSQLByteParser(b.getvalue())
    v0 = time.clock() - t
    rat_sum = 0.0
    for (r,c,rat) in pb.read_all_triples():
        counter += 1
        rat_sum += rat
    v = time.clock()
    print "[triples] Counter={0} timer={1}s tup/s={2} build={3} ratsum={4}".format(counter,v, counter/v, v0, rat_sum)

def test3():
    t = time.clock()
    conn = psycopg2.connect('dbname=chrisre')
    cur = conn.cursor()
    b = cStringIO.StringIO()
    cur.copy_expert("COPY (SELECT UNNEST(L) FROM lr_model_instance WHERE mid=11) TO STDOUT WITH BINARY", b)
    v = time.clock() - t
    print "Time to execute copy query = {0}".format(v)

    t = time.clock()
    counter = 0
    pb = PSQLByteParser(b.getvalue())
    v0 = time.clock() - t
    rat_sum = 0.0
    for (r,c,rat) in pb.read_all_2darrays():
        counter += 1
        rat_sum += rat
    v = time.clock()
    print "[triples] Counter={0} timer={1}s tup/s={2} build={3} ratsum={4}".format(counter,v, counter/v, v0, rat_sum)
