"""
This module contains generic utilities that can be used
by any model.
"""

from math import *

def grad(model, w, b):
    """
    Calculates the gradient
    @type model: vector
    @param model: model
    @type w: vector
    @param w: model vector
    @type b: double
    @param b: bias
    
    @rtype: vector
    @return: gradient
    """
    n = min(len(model),len(w)) 
    err = sum([model[i]*w[i] for i in range(n)]) - b
    return [ wi * err for wi in w]

def dot(u,v):
    """
    Computes and returns the dot product of given two vectors
    @type u: vector
    @param u: vector1
    @type v: vector
    @param v: vector2

    @rtype: double
    @return: dot product value of the vectors
    """
    assert(len(u) == len(v))
    return sum([u[i]*v[i] for i in range(len(u))])

def grad(w,b,x,y):    
    err = dot(w,x) + b - y
    return (scale(w,err), err*b)
    

def scale(w, b):
    """
    Scales the given vector with the given scale factor and return it
    @type w: vector
    @param w: vector to be scaled
    @type b: double
    @param b: scale factor
    
    @rtype: vector
    @return: scaled vector
    """
    return [wi*b for wi in w]

def scale_and_add(x,y,s):
    """
    Scales given vector y and add it to the x and return it
    
    @type x: vector
    @param x: vector to be added with the scaled one
    @type y: vector
    @param y: vector to be scaled
    @type s: double
    @param s: scale factor

    @rtype: vector
    @return: x + y * s
    """
    n = len(x)
    assert( n == len(y))
    return [x[i] + y[i]*s for i in range(n)] 

def dot_ds(dx, sy):
    """
    Computes and returns dot product of two vectors (at least one of them is sparse)
    @type dx: vector
    @param dx: first component of the dot product
    @type sy: vector
    @param sy: second component of the dot product
   
    @rtype: double
    @return: dot product of dx and sy
    """
    s = 0.0
    for (k,v) in sy:
        s = s + dx[k]*v
    return

def sigma(v):
    """
    Returns sigma valu
    
    @type v: double
    @param v: v

    @rtype: double
    @return: sigma value of v
    """
    if(v > 30):
        return 1.0/(1.0 + exp(-v))
    else:
        return exp(v)/(1+exp(v))
################33
# inplace
#
def scale_i(x,c):
    """
    Scales the given vector
    @type x: vector
    @param x: vector to be scaled
    @type c: double
    @param c: scale factor
    """
    for i in range(len(x)):
        x[i] = x[i] * c

def scale_and_add_i(x, y, c):
    """
    Scales y and adds it to x
    x += y * c

    @type x: vector
    @param x: vector to be modified
    @type y: vector
    @param y: vector to be scaled and added to x
    @type c: double
    @param c: scale factor

    """
    n = min(len(x), len(y))
    for i in range(n):
        x[i] = x[i] + y[i] * c
        
def scale_and_add_ds(dx, sz, c):
    """
    Scales sz (which is sparse) and adds it to x
    dx += sz * c

    @type dx: vector
    @param dx: vector to be modified
    @type sz: vector
    @param sz: vector to be scaled and added to dx
    @type c: double
    @param c: scale factor
    """
    for (k,v) in sz:
        dx[k] = dx[k] + v * c

## splits    
def dot_dss(w, ks, vs):
    """
    Calculates and returns dot products of w and vs
    @type w: vector
    @param w: first component of the dot product
    @type ks: vector
    @param ks: indexes of the sparse vector
    @type vs: vector
    @param vs: values of the sparse vector
    
    @rtype: double
    @return: dot product of two vectors
    """
    s = 0
    for i in range(len(ks)):
        s = s + w[ks[i]]*vs[i]
    return s

def scale_and_add_dss(w, index, vectors, c):
    """
    Scales vectors and adds it to w

    @type w: vector
    @param w: vector to be modified
    @type index: vector
    @param index: indexes of the sparse vector
    @type vectors: vector
    @param vectors: values of the sparse vector
    @type c: double
    @param c: scale factor
    """
    for i in range(len(index)):
        w[index[i]] = w[index[i]] + vectors[i]*c

#### PROJECTIONS
def l2_project(x, B):
    """
    l2 project function
    @type x: vector
    @param x: x
    @type B: double 
    @param B: B
    """
    norm_sq = sum([xi*xi for xi in x])
    if norm_sq > B*B:
        norm = sqrt(norm_sq)
        for i in range(len(x)):
            x[i] = x[i]*B/norm



# Prox temrs
def l1_shrink(x,u):
    """
    l1 shrink function
    @type x: vector
    @param x: x
    @type u: double 
    @param u: u
    """
    for i in range(len(x)):
        if x[i] > u:
            x[i] = x[i] - u
        else:
            if x[i] < -u:
                x[i] = x[i] + u
            else:
                x[i] = 0
            

def l1_shrink_mask(x,u,indexes):
    """
    l1 shrink function
    @type x: vector
    @param x: x
    @type u: double 
    @param u: B
    @type indexes: vector
    @param indexes: indexes of the sparse vector
    """
    for i in indexes:
        if x[i] > u:
            x[i] = x[i] - u
        else:
            if x[i] < -u:
                x[i] = x[i] + u
            else:
                x[i] = 0
            


# helper for models                
def incremental_average(w_hat, w, steps):
    """
    Incremental average function
    @type w_hat: vector
    @param w_hat: w hat
    @type w: vector
    @param w: model vector
    @type steps: number
    @param steps: steps
    """
    if steps == 0:
        for i in range(len(w_hat)):
            w_hat[i] = w[i]
    else:
        steps = float(steps)
        for i in range(len(w_hat)):
            w_hat[i] = w[i] * 1.0/steps + steps/(steps + 1) * w_hat[i]

    
