#include "postgres.h"
#include "utils/memutils.h"
#include "utils/typcache.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include "funcapi.h"
#include "init_helper.h"
#include <math.h>

#include "access/heapam.h"
#include "catalog/namespace.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_type.h"
#include "funcapi.h"
#include "nodes/nodeFuncs.h"
#include "parser/parse_coerce.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/syscache.h"
#include "utils/typcache.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


PG_FUNCTION_INFO_V1(factor_partition);

#ifndef min
#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif

static void* gd = NULL;

char names[2][10] = {"row_bucket", "col_bucket"};
int oids[2] = {23, 23};
int typmods[2] = {-1, -1};

void shuffle_array(int *arr, int dim) {
  int i;
  for(i = 0; i < dim; i ++) {
    int rand_ind = rand() % dim;
    int temp = arr[rand_ind];
    arr[rand_ind] = arr[i];
    arr[i] = temp;
  }
}




static void
internal_get_result_type(Oid funcid,
			 Node *call_expr,
			 ReturnSetInfo *rsinfo,
			 Oid *resultTypeId,
			 TupleDesc *resultTupleDesc)
{
  HeapTuple	tp;
  Form_pg_proc procform;
  Oid			rettype;
  TupleDesc	tupdesc;

  /* /\* First fetch the function's pg_proc row to inspect its rettype *\/ */
  /* tp = SearchSysCache1(PROCOID, ObjectIdGetDatum(funcid)); */
  /* if (!HeapTupleIsValid(tp)) */
  /*   elog(ERROR, "cache lookup failed for function %u", funcid); */
  /* procform = (Form_pg_proc) GETSTRUCT(tp); */

  /* rettype = procform->prorettype; */
  /* elog(INFO, "rettype: %d\n", rettype); */

  if (resultTupleDesc) {
    *resultTupleDesc = lookup_rowtype_tupdesc_copy(rettype, -1);
  }

  /* ReleaseSysCache(tp); */
}

Datum
factor_partition(PG_FUNCTION_ARGS)
{
  Datum values[2];
  int tuplen;
  bool *nulls;
  HeapTuple tuple;
  TupleDesc tupdesc;

  int nRows         = PG_GETARG_INT32(0);
  int nCols         = PG_GETARG_INT32(1);
  int seed          = PG_GETARG_FLOAT8(2);
  int nSplits2      = PG_GETARG_INT32(3);
  HeapTupleHeader t = PG_GETARG_HEAPTUPLEHEADER(4);

  
  gd = check_init(gd);
  int *seed_val, *row_perm, *col_perm, i;
  seed_val = get_key(gd, "seed");
  
  if(seed_val == NULL || (seed_val != NULL && *seed_val != seed) ||
     !MemoryContextContains(TopTransactionContext, (void*) seed_val)) {
    seed_val = MemoryContextAlloc(TopTransactionContext, sizeof(int) * 1);
    row_perm = MemoryContextAlloc(TopTransactionContext, sizeof(int) * nRows);
    col_perm = MemoryContextAlloc(TopTransactionContext, sizeof(int) * nCols);

    *seed_val = seed;
    for(i = 0; i < nRows; i ++)
      row_perm[i] = i;
    for(i = 0; i < nCols; i ++)
      col_perm[i] = i;

    srand(seed);
    shuffle_array(row_perm, nRows);
    shuffle_array(col_perm, nCols);

    set_key(gd, "seed", seed_val);
    set_key(gd, "row_perm", row_perm);
    set_key(gd, "col_perm", col_perm);
  }

  row_perm = (int *) get_key(gd, "row_perm");
  col_perm = (int *) get_key(gd, "col_perm");

  /* internal_get_result_type(fcinfo->flinfo->fn_oid, */
  /* 			   fcinfo->flinfo->fn_expr, */
  /* 			   (ReturnSetInfo *) fcinfo->resultinfo, */
  /* 			   NULL, */
  /* 			   &tupdesc); */
  tupdesc = lookup_rowtype_tupdesc_copy(9385829, -1);

  BlessTupleDesc( tupdesc );

  int nSplits = (int) sqrt(nSplits2);

  bool isnull;
  int r0 = DatumGetInt32(GetAttributeByNum(t, 1, &isnull)) - 1;
  int c0 = DatumGetInt32(GetAttributeByNum(t, 2, &isnull)) - 1;

  if (r0 >= nRows || c0 >= nCols || r0 < 0 || c0 < 0)
    elog(ERROR, "r0: %d, c0: %d", r0, c0);

  int r = row_perm[r0];
  int c = col_perm[c0];

  int r_val1 = ((int)((double) r / ((double) nRows / (double) nSplits)));
  int val2 = nSplits - 1;
  int c_val1 = ((int)((double) c / ((double) nCols / (double) nSplits)));

  int row_bucket = min(r_val1, val2);
  int col_bucket = min(c_val1, val2);

  values[0] = Int32GetDatum( row_bucket );
  values[1] = Int32GetDatum( col_bucket );
  tuplen = tupdesc->natts;

  nulls = palloc( tuplen * sizeof( bool ) );

  /* // build tuple from datum array */
  tuple = heap_form_tuple( tupdesc, values, nulls );

  pfree( nulls );

  // A float8 datum palloc's space, so if we free them too soon,
  // their values will be corrupted (so don't pfree here, let
  // PostgreSQL take care of it.)
  // pfree(values);
   
  PG_RETURN_DATUM( HeapTupleGetDatum( tuple ) );
}
