/*
 * testlibpq.c
 *
 *      Test the C version of libpq, the PostgreSQL frontend library.
 */
#include <string>
#include <sstream>
#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <libpq-fe.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <iostream>
#include "postgres_util.h"

static void
exit_nicely(PGconn *conn)
{
  PQfinish(conn);
  exit(1);
}


#define FETCH_SIZE 20000
#define COPY_BUF_LIMIT 100000

int get_int(const char *ptr) {
  return ntohl(*((uint32_t *) ptr));
}

std::string get_split_name(std::string output_name, int id) {
  std::ostringstream output;
  output << output_name << "_split_" << id;
  return output.str();
}

void check_result(PGconn *conn, PGresult *res) {
  if (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK)
    {
      fprintf(stderr, "query failed: %s", PQerrorMessage(conn));
      PQclear(res);
      exit_nicely(conn);
    }
}

uint32_t pack_int32(int val) {
  return htonl(val);
}

void permute_and_slice_data(std::string db_name, std::string partition_fnc, std::string input_view_name,
			    std::string output_base_name, int nSplits) {
  PGconn     *conn;
  PGresult   *res;
  std::string conninfo = "dbname = " + db_name;

  /* Make a connection to the database */
  conn = PQconnectdb(conninfo.c_str());

  /* Check to see that the backend connection was successfully made */
  if (PQstatus(conn) != CONNECTION_OK)
    {
      fprintf(stderr, "Connection to database failed: %s",
	      PQerrorMessage(conn));
      exit_nicely(conn);
    }

  res = PQexec(conn, "select library_setup()");
  check_result(conn, res);

  std::vector<std::string> split_table_names;
  for(int i = 0; i < nSplits; i ++)
    split_table_names.push_back(get_split_name(output_base_name, i));

  //drop statements
  std::ostringstream drop_statements;
  for(int i = 0; i < nSplits; i ++) {
    drop_statements << "DROP TABLE IF EXISTS " << split_table_names[i] << ";\n";
  }

  res = PQexec(conn, drop_statements.str().c_str());
  check_result(conn, res);

  //create statements
  std::ostringstream create_statements;
  for(int i = 0; i < nSplits; i ++) {
    create_statements << "SELECT * INTO " << split_table_names[i] << " FROM " << input_view_name << " as x LIMIT 0;\n";
  }

  res = PQexec(conn, create_statements.str().c_str());
  check_result(conn, res);

  
  //std::string partition_query = "SELECT " + partition_fnc + "(t.*), * FROM " + input_view_name + " as t;";
  std::string partition_query = "SELECT " + partition_fnc + "(t.*), * FROM " + input_view_name + " as t;";
  //std::string partition_query = "SELECT ((row * 10 + col * 23) % 7), * FROM " + input_view_name + " as t;";
  std::string cursor_query = "DECLARE victor_cursor_" + output_base_name + " BINARY CURSOR FOR " + partition_query;
  res = PQexec(conn, "BEGIN");
  check_result(conn, res);

  res = PQexec(conn, cursor_query.c_str());
  check_result(conn, res);

  std::ostringstream retrieve_query;
  retrieve_query << "FETCH " << FETCH_SIZE << " in victor_cursor_" << output_base_name;

  res = PQexec(conn, retrieve_query.str().c_str());
  check_result(conn, res);
  
  int nFields = PQnfields(res);
  int num_of_tuples = PQntuples(res);
  uint16_t packed_fields = htons(nFields - 1);
  uint16_t *packed_fields_ptr = &packed_fields;

  std::vector<Victor_Database *> vdbs;
  for(int i = 0; i < nSplits; i ++) {
    vdbs.push_back(new Victor_Database(conninfo.c_str(), split_table_names[i], COPY_BUF_LIMIT));
  }
  
  while(num_of_tuples > 0) {
    for (int i = 0; i < num_of_tuples; i ++) {
      int partition_id = get_int(PQgetvalue(res, i, 0));
      Victor_Database *vdb = vdbs[partition_id];
      vdb->copy_to_table((char *) packed_fields_ptr, 2);
      for (int j = 1; j < nFields; j ++) {
	vdb->copy_field_to_table(PQgetvalue(res, i, j), PQgetlength(res, i, j));
      }
      vdb->check_flush();
    }
    res = PQexec(conn, retrieve_query.str().c_str());
    check_result(conn, res);
    num_of_tuples = PQntuples(res);
  }

  for(int i = 0; i < nSplits; i ++)
    vdbs[i]->flush();

  res = PQexec(conn, "END");
  PQclear(res);

  /* close the connection to the database and cleanup */
  PQfinish(conn);
}


int
main(int argc, char **argv)
{
  permute_and_slice_data("test_victor", "c_partition", "movielens1m", "c_version_test", 7);

  return 0;
}
