import random, struct, cStringIO, StringIO
import psycopg2
import time
import parse_fast

#################################################################
## This module build the copy internals for the factored modules
## One key feature: it internally buffers the cursors so that the
## The RBDMS takes care of all the memory management.
## For speed, the internals are CYTHON. The frontend is pure python.
#################################################################

global_debug_flag = False
# Binary Parsers used to translate network byte orders
# to database native byteorders
format_string = '>hiiiiid'
struct_triple_parser = struct.Struct(format_string)
psql_header = b'PGCOPY\n\377\r\n\0\0\0\0\0\0\0\0\0'
real_psql_header = bytearray(b'PGCOPY\n') + bytearray(b'\377\r\n\0')
struct_int32  = struct.Struct('>i')
struct_int64  = struct.Struct('>q')
struct_int16  = struct.Struct('>h')
struct_double = struct.Struct('>d')
double_oid    = 701 # see parse_fast.pyx

# A special parser to build the binary copy header
string_struct        = struct.Struct('19p') # 19 bytes
size_int             = 4
size_double          = 8
size_expression      = struct.calcsize(format_string)

#
model_split_query  = "DROP TABLE IF EXISTS {0}; CREATE TABLE {0} (id int, vec float8[]);"

# No longer used.
def write_triple_inplace(b,off, t):
    (row,col,rating) = t
    struct_triple_parser.pack_into( b, off, 3, size_int, row, size_int, col, size_double, rating)
    return off + size_expression

# fast triple packing and writing: (row, col, value) into a buffer b
cdef int write_triple(b,t):
    cdef int row, col
    cdef double rating
    (row,col,rating) = t
    b.write(struct_triple_parser.pack( 3, size_int, row, size_int, col, size_double, rating))
    return size_expression

# write the PSQL end-of-stream character (eos)
cdef write_eos(b):
    b.write( struct_int16.pack( -1 ) ) 

# Compute the bucket a particular row (or column) goes into
# given a current bucket size. Range saved with slices
cdef get_bucket(int r, int bucket_size, int nSlices):
    return max(0, min(r/bucket_size, nSlices - 1))

# A special class to handle triple flushes to the RDBMS
class OutputBuffer:
    def __init__(self, cur, table_name, nBuf, debug=global_debug_flag, writeHeader=True):
        self.cur   = cur
        self.debug = debug
        self.nBuf  = nBuf
        self.query = "COPY {0} FROM STDIN WITH BINARY;".format(table_name) 
        self.writeHeader = writeHeader
        self.init_buffer()

    # Init sets up the buffer, writes out the psql header for COPY
    def init_buffer(self):
        self.buffer = cStringIO.StringIO()
        if self.writeHeader:
            self.buffer.write(psql_header)
        self.index  = 0

    # Writes the end of stream and then COPYs the data in binary
    def flush(self):
        write_eos(self.buffer)
        if self.debug:
            print "\t[Flushing Buffer] for \"{0}\"".format(self.query)
        # This tmp hack shouldn't be necessary... 
        tmp = cStringIO.StringIO(self.buffer.getvalue())
        self.cur.copy_expert(self.query, tmp)
        tmp.close()

    # flush, then free all the resources
    def close(self):
        self.flush()
        self.buffer.close()
        
    # Write tuple
    def internal_write_triple(self, t):
        self.index += write_triple(self.buffer, t) 

    # Output a Triple
    def check_flush( self ):
       if self.index > self.nBuf:
           self.flush()
           self.buffer.close()
           self.init_buffer()
        

    def internal_write_double_array( self, rid, a):
        # describe the tuple
        self.buffer.write( struct_int16.pack( 2 ) ) # nFields
        # Write the id
        self.buffer.write( struct_int32.pack( size_int ) )
        self.buffer.write( struct_int32.pack( rid ) )
        
        # Describe the array
        self.buffer.write( struct_int32.pack( size_int * 5 + len(a) * (size_int + size_double) ) ) # Field size        
        self.buffer.write( struct_int32.pack( 1 ) ) # write ndims
        self.buffer.write( struct_int32.pack( 0 ) ) # has_null
        self.buffer.write( struct_int32.pack ( double_oid ) ) # element type
        # Now describ the array
        self.buffer.write( struct_int32.pack( len(a) ) ) #length        
        self.buffer.write( struct_int32.pack( 1 ) ) # lbound
        for d in a:
            self.buffer.write( struct_int32.pack( size_double ) )
            self.buffer.write( struct_double.pack( d ) )

    # These functions form the intended interface
    def output_triple( self, t ):
       self.internal_write_triple( t )
       self.check_flush()

    #MISNAMED! Should be (int, array)
    def output_double_array( self, rid, a ):
        self.internal_write_double_array(rid, a)
        self.check_flush()

    def write_int32( self, i):
        self.buffer.write( struct_int32.pack( i ) )

    def write_double( self, d):
        self.buffer.write( struct_double.pack( d ) )

    # TODO: Add the dual read method.
    def output_lr_model( self, mid,L, R, moviemean, stepsize, B):
        # write # of fields
        self.buffer.write( struct_int16.pack(6) )

        # Write mid
        self.write_int32( size_int ) 
        self.write_int32( mid )
        
        # Write L
        rows      = len(L)
        rank      = len(L[0])
        # array header is 3 ints.
        # Then 4 ints to describe the bounds
        L_size    = size_int * (3 + 4) + rows*rank * (size_int + size_double)
        ndims     = 2
        has_nulls = 0
        for x in [L_size, ndims, has_nulls, double_oid]:
            self.write_int32 ( x )
        
        # Describe the array bounds
        for x in [ rows, 1, rank, 1 ]:
            self.write_int32 ( x )
        # write the elements
        for l in L:
            for x in l:
                self.write_int32( size_double )
                self.write_double(  x )
                
        # write R
        cols    = len(R)
        rank    = len(R[0])
        R_size  = size_int * (3 + 4) + cols*rank * (size_int + size_double)
        ndims   = 2
        has_nulls = 0
        for x in [R_size, ndims, has_nulls, double_oid]:
            self.write_int32 ( x )            
        # Describe the array bounds
        for x in [ cols, 1, rank, 1 ]:
            self.write_int32 ( x )
        # write the elements
        for r in R:
            for x in r:
                self.write_int32( size_double )
                self.write_double( x )
                
        # write movie mean, stepsize, and B
        for x in [ moviemean, stepsize, B ]:
            self.write_int32( size_double )
            self.write_double ( x ) 
        if global_debug_flag:
            print "\t Storing rows={0} cols={1} rank={2}".format(rows, cols, rank)

####
            

cdef internal_retrieve_sparse_model_simple(dbname, int mid):
    import low_rank_helper
    q = "SELECT array_upper(L,1) as nRows, array_upper(L,2) as l_rank, array_upper(R,1) as ncols, array_upper(R,2) as r_rank, moviemean as moviemean, stepsize as stepsize, b as b FROM lr_model_instance WHERE mid={0}".format(mid)
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()

    cur.execute(q)
    print "[irsm] q={0}".format(q)
    (nrows, l_rank, ncols, r_rank, moviemean, stepsize, b) = cur.fetchone()

    # Parse L
    buf = cStringIO.StringIO()
    cur.copy_expert("COPY (select L,R from lr_model_instance where mid={0}) TO STDOUT WITH BINARY".format(mid), buf)
    pb = parse_fast.PSQLByteParser(buf.getvalue())

    (L,R) = pb.read_single_pair_2d()
    buf.close()
    
    return [L,R,moviemean,stepsize,b,b*b]

           
# This is responsible for partitioning the data
# TODO: Add in a description of the data.
cdef internal_partition_data(dbname,seed, query_str, int nSlices, int epoch, int mid, int asize):
    cdef int msax_cols, max_rows, max_rank, row_bucket_size, col_bucket_size
    cdef double t
    victor_name    = "victor_{0}_{1}".format(mid,epoch)
    ssl_model_name = "{0}_lmodel".format(victor_name)
    ssr_model_name = "{0}_rmodel".format(victor_name)
    partition_prefix = victor_name
    
    stats_query    = "SELECT array_upper(R,1) as max_cols, array_upper(L,1) as max_rows, array_upper(L,2) as max_rank from lr_model_instance where mid = {0}".format(mid)
    
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()
    cur.execute(stats_query)
    (max_cols, max_rows, max_rank) = cur.fetchone()
    
    row_perm = range(max_rows)
    col_perm = range(max_cols)
    # now do the permutation
    # the order is critical here!
    t = time.time()
    random.seed(seed)
    random.shuffle(row_perm)
    random.shuffle(col_perm)
    if global_debug_flag:
        print "\t Finished Permutations in {0} for {1} rows and {2} cols".format(time.time() - t, max_rows, max_cols)
    
    # Create the relevant tables
    t = time.time()
    output_buffers = []
    for i in xrange(nSlices * nSlices):
        table_name = "{0}_data_{1}".format(partition_prefix, i)
        query      = "DROP TABLE IF EXISTS {0}; CREATE TABLE {0} (row int, col int, rating float8);".format(table_name)
        output_buffers.append( OutputBuffer(cur, table_name, asize) ) 
        cur.execute(query)

    if global_debug_flag:
        print "\t Creating Tables took {0}s".format(time.time() - t)
    
    # Partition the data
    row_bucket_size = max_rows/nSlices
    col_bucket_size = max_cols/nSlices

    t = time.time()
    # Any name for the cursor makes it server side, which is what we
    # want. Although we set arraysize here, pyscopg2 doesn't listen
    # (you can see it with the FETCH 1 messages so we have to remember
    # to add the buffer size we want with each fetchmany().
    cur            = conn.cursor('factor_parser_cursor_chicken') 
    cur.arraysize  = asize
    cur.execute( query_str )
    if global_debug_flag:
        print "\t Cursor execute completed {0}".format(time.time() - t)

    cdef int row, col
    cdef double rating
    cdef int new_row, new_col, row_bucket, col_bucket, zz
    
    if global_debug_flag:
        print "\t Cursor requesting {0} items per iteration".format(asize)
    results = cur.fetchmany(asize)
    while results:
        for (row,col,rating,) in results:
            new_row    = row_perm[ row - 1 ]
            new_col    = col_perm[ col - 1 ]
            row_bucket = get_bucket(new_row, row_bucket_size, nSlices)
            col_bucket = get_bucket(new_col, col_bucket_size, nSlices)
            zz         = row_bucket*nSlices + col_bucket
            if(nSlices == 1):
                zz = 0
            output_buffers[zz].output_triple( (row,col,rating) ) 
        results = cur.fetchmany(asize)

    # flush the remaining buffers
    for i in xrange(nSlices*nSlices):
        output_buffers[i].close()
        
    cur.close()
    conn.commit()
    conn.close()

    if global_debug_flag:
        print "\t Finished in {0}s".format(time.time() - t)

# TODO: Fix the partition model
cdef internal_partition_model(dbname, int seed, query_str, int nSlices, int epoch, int mid, int asize):
    victor_name    = "victor_{0}_{1}".format(mid,epoch)
    ssl_model_name = "{0}_lmodel".format(victor_name)
    ssr_model_name = "{0}_rmodel".format(victor_name)
    partition_prefix = victor_name

    stats_query    = "SELECT array_upper(R,1) as max_cols, array_upper(L,1) as max_rows, array_upper(L,2) as max_rank from lr_model_instance where mid = {0}".format(mid)
    
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()
    cur.execute(stats_query)
    (max_cols, max_rows, max_rank) = cur.fetchone()

    row_perm = range(max_rows)
    col_perm = range(max_cols)
    # now do the permutation
    # the order is critical here!
    t = time.time()
    random.seed(seed)
    random.shuffle(row_perm)
    random.shuffle(col_perm)
    if global_debug_flag:
        print "\t [Partition Model] Finished Permutations in {0} for {1} rows and {2} cols".format(time.time() - t, max_rows, max_cols)
    

    # Partition the data
    row_bucket_size = max_rows/nSlices
    col_bucket_size = max_cols/nSlices

    [L,R,moviemean,stepsize,b,b2] = internal_retrieve_sparse_model_simple(dbname, mid)

    ## Build the L split tables and the output bucket
    output_bucket = []
    for slice_no in xrange(nSlices):
        table_name = "{0}{1}".format(ssl_model_name, slice_no)
        query      = model_split_query.format(table_name)
        output_bucket.append ( OutputBuffer(cur, table_name, asize) )
        cur.execute ( query )


    for i in xrange(max_rows):
        output_bucket[get_bucket(row_perm[i], row_bucket_size, nSlices)].output_double_array(i, L[i])    

    for ob in output_bucket:
        ob.close()
    
    output_bucket = []
    for slice_no in xrange(nSlices):
        table_name = "{0}{1}".format(ssr_model_name, slice_no)
        query      = model_split_query.format(table_name)
        output_bucket.append ( OutputBuffer(cur, table_name, asize) )
        cur.execute ( query )        
    
    for j in xrange(max_cols):
        output_bucket[get_bucket(col_perm[j], col_bucket_size, nSlices)].output_double_array(j, R[j])
    
    for ob in output_bucket:
        ob.close()

    cur.close()
    conn.commit()
    conn.close()

cdef internal_retrieve_sparse_model(dbname, int mid, int epoch_sign, int nsplits, int round, int chunk):
    import low_rank_helper
    q = "SELECT array_upper(L,1) as nRows, array_upper(L,2) as l_rank, array_upper(R,1) as ncols, array_upper(R,2) as r_rank, moviemean as moviemean, stepsize as stepsize, b as b FROM lr_model_instance WHERE mid={0}".format(mid)
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()
    cur.execute(q)
    (nrows, l_rank, ncols, r_rank, moviemean, stepsize, b) = cur.fetchone()

    # Parse L
    l_name = low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))
    buf = cStringIO.StringIO()
    cur.copy_expert("COPY {0} TO STDOUT WITH BINARY".format(l_name), buf)
    pb = parse_fast.PSQLByteParser(buf.getvalue())

    # NB: We want L to be sparse!
    # we use the fact that len(L[i]) == 0 tells us that we are not responsible for that row.
    L = [ [] for i in xrange(nrows)]
    for (i,l) in pb.read_all_ia_pairs():
        L[i] = l
    buf.close()

    r_name = low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))
    buf = cStringIO.StringIO()
    cur.copy_expert("COPY {0} TO STDOUT WITH BINARY".format(r_name), buf)
    pb = parse_fast.PSQLByteParser(buf.getvalue())
    R = [ [] for i in xrange(ncols)]
    
    for (i,r) in pb.read_all_ia_pairs():
        R[i] = r
    buf.close()

    for_sd = (mid, epoch_sign, nsplits, round, chunk)
    return ([L,R,moviemean,stepsize,b,b*b], for_sd)


cdef internal_update_sparse_segment(dbname, int mid, int epoch_sign, int nsplits, int round, int chunk, L,R, double moviemean,double stepsize, double B, double B2, int asize):
    import low_rank_helper
    import psycopg2
    cdef int i
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()

    l_name = low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))
    cur.execute("DELETE FROM {0};".format(l_name))
    cur    = conn.cursor()
    
    ob = OutputBuffer(cur, l_name, asize)
    for i in xrange(len(L)):
        if len(L[i]) > 0:
            ob.output_double_array(i, L[i])
    ob.close()

    r_name = low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))
    cur.execute("DELETE FROM {0};".format(r_name))
    ob = OutputBuffer(cur, r_name, asize)
    for i in xrange(len(R)):
        if len(R[i]) > 0:
            ob.output_double_array(i, R[i])
    ob.close()
    cur.close()
    conn.commit()
    conn.close()


cdef internal_store_model(dbname, int mid, L, R, double moviemean, double stepsize, double B, int asize):
    table_name = "lr_model_instance"
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur    = conn.cursor()
    cur.execute("delete from {0} where mid={1}".format(table_name, mid))
    
    ob = OutputBuffer(cur, table_name, asize)
    ob.output_lr_model(mid, L,R, moviemean, stepsize, B)
    ob.close()
    
    cur.close()
    conn.commit()
    conn.close()

def internal_rebuild_model(dbname, int mid, int epoch_sign, int nsplits, int asize):
    from low_rank_helper import tablename_lmodel, tablename_rmodel
    cdef int i
    cdef double moviemean, stepsize, B
    
    table_name     = "lr_model_instance" 
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()
    # TODO: write a function
    stats_query    = "SELECT array_upper(R,1) as max_cols, array_upper(L,1) as max_rows, array_upper(L,2) as max_rank, moviemean, stepsize, b from lr_model_instance where mid = {0};".format(mid)
    cur.execute(stats_query)
    if global_debug_flag:
        print "stats_query = {0}".format(stats_query)
    (num_cols, num_rows, max_rank, moviemean, stepsize, B) = cur.fetchone()
    
    L   = [[0.0]*max_rank for i in xrange(num_rows)]
    R   = [[0.0]*max_rank for i in xrange(num_cols)]
        
    for i in xrange(nsplits):
        b = cStringIO.StringIO()
        copy_query = "COPY {0}(id,vec) TO STDOUT WITH BINARY;".format(tablename_lmodel(mid, epoch_sign, i))
        if global_debug_flag:
            print "\t [L] copy_query = {0}".format(copy_query)
        cur.copy_expert(copy_query, b)
        pb = parse_fast.PSQLByteParser(b.getvalue())
        for (i, x) in pb.read_all_ia_pairs():
            L[i] = x
        b.close()

    for i in xrange(nsplits):
        b = cStringIO.StringIO()
        copy_query = "COPY {0}(id,vec) TO STDOUT WITH BINARY;".format(tablename_rmodel(mid, epoch_sign, i))
        if global_debug_flag:
            print "\t [R] copy_query = {0}".format(copy_query)
        cur.copy_expert(copy_query, b)
        pb = parse_fast.PSQLByteParser(b.getvalue())
        for (i, x) in pb.read_all_ia_pairs():
            R[i] = x
        b.close()

    cur.execute("delete from {0} where mid={1}".format(table_name, mid))
    ob = OutputBuffer(cur, table_name, asize)
    ob.output_lr_model(mid, L, R, moviemean, stepsize, B)
    ob.close()
    cur.close()
    
    conn.commit()
    conn.close()

    
###########################################################################
# These are the front end for the permute and slice functions.
default_asize = 2000000
def permute_and_slice_data(dbname,seed, query_str, nSlices, epoch, mid,   asize=default_asize):
    internal_partition_data(dbname,seed, query_str, nSlices, epoch, mid, asize)

def permute_and_slice_model(dbname, seed, query_str, nSlices, epoch, mid, asize=default_asize):
    internal_partition_model(dbname, seed, query_str, nSlices, epoch, mid, asize)

## Sparse model functions.

def retrieve_sparse_model(dbname, mid, epoch_sign, nsplits, round, chunk):
    return internal_retrieve_sparse_model(dbname, mid, epoch_sign, nsplits, round, chunk)

def update_sparse_segment(dbname, mid, epoch_sign, nsplits, round, chunk, L, R, moviemean, stepsize, B, B2, asize=default_asize):
    internal_update_sparse_segment(dbname, mid, epoch_sign, nsplits, round, chunk, L, R, moviemean, stepsize, B, B2,asize)

## model store
def store_model(dbname, mid, L, R, moviemean, stepsize, B, B2, asize=default_asize):
    internal_store_model(dbname, mid, L, R, moviemean, stepsize, B, asize)

## rebuild the spread out model
def sparse_rebuild_model(dbname, mid, epoch_sign, nsplits, asize=default_asize):
    internal_rebuild_model(dbname, mid, epoch_sign, nsplits, asize)

def retrieve_sparse_model_simple(dbname, mid):
    return internal_retrieve_sparse_model_simple(dbname,mid)
