"""
This module build the copy internals for the factored modules
One key feature: it internally buffers the cursors so that the
The RBDMS takes care of all the memory management.
For speed, the internals are CYTHON. The frontend is pure python.
"""

import random, struct, cStringIO, StringIO
import psycopg2
import time
import src.vsql_core.utils.factored_parser.parse_fast
from psql_util import *

import mmap_wrapper

global_debug_flag = False            

dir_base = '/scratch.1/VICTOR_SQL'

model_simple_base = "model_simple"
model_l_r_base = "model_l_r"
stats_simple_base = "stats_simple"
stats_base = "stats"

"""
Compute the bucket a particular row (or column) goes into
given a current bucket size. Range saved with slices
"""
cdef get_bucket(int r, int bucket_size, int nSlices):
    return max(0, min(r/bucket_size, nSlices - 1))

cdef get_VictorFile(file_name, schema, read_or_write_mode):
    try:
        #File first needs to be written at least 1 byte
        #If the file doesnt exist, we create it
        f = open(file_name, "r+b")
    except IOError:
        f = open(file_name, "wb")
        f.write("t")
        f.close()
        f = open(file_name, "r+b")
    mem_map = mmap_wrapper.MMapWrapper(f.fileno(), 0, read_or_write_mode)
    return VictorFile(mem_map, schema, read_or_write_mode)

cdef writeSingleTupleToFile(file_name, schema, t):
    vdb = get_VictorFile(file_name, schema, mmap_wrapper.WRITE_MODE)
    vdb.write_tuples(t)
    vdb.flush()

cdef internal_retrieve_sparse_model_simple(dbname, int mid):
    import low_rank_helper
    vdb = get_VictorFile("{0}/{1}{2}.dat".format(dir_base, model_simple_base, mid),
                         [int4oid, int4oid, int4oid, int4oid, float8oid, float8oid, float8oid],
                         mmap_wrapper.READ_MODE)
    (nrows, l_rank, ncols, r_rank, moviemean, stepsize, b) = vdb.read_tuples()[0]

    vdb = get_VictorFile("{0}/{1}{2}.dat".format(dir_base, model_l_r_base, mid),
                         [float82darrayoid, float82darrayoid],
                         mmap_wrapper.READ_MODE)
    (L, R) = vdb.read_tuples()[0]    
    
    return [L,R,moviemean,stepsize,b,b*b]

           
# This is responsible for partitioning the data
# TODO: Add in a description of the data.
cdef internal_partition_data(dbname,seed, query_str, int nSlices, int epoch, int mid, int asize):
    cdef int msax_cols, max_rows, max_rank, row_bucket_size, col_bucket_size
    cdef double t
    victor_name    = "victor_{0}_{1}".format(mid,epoch)
    ssl_model_name = "{0}_lmodel".format(victor_name)
    ssr_model_name = "{0}_rmodel".format(victor_name)
    partition_prefix = victor_name
    
    stats_query    = "SELECT array_upper(R,1) as max_cols, array_upper(L,1) as max_rows, array_upper(L,2) as max_rank from lr_model_instance where mid = {0}".format(mid)
    
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()
    cur.execute(stats_query)
    (max_cols, max_rows, max_rank) = cur.fetchone()
    
    row_perm = range(max_rows)
    col_perm = range(max_cols)
    # now do the permutation
    # the order is critical here!
    t = time.time()
    random.seed(seed)
    random.shuffle(row_perm)
    random.shuffle(col_perm)
    if global_debug_flag:
        print "\t Finished Permutations in {0} for {1} rows and {2} cols".format(time.time() - t, max_rows, max_cols)
    
    # Create the relevant tables
    t = time.time()
    output_vdbs = []

    for i in xrange(nSlices * nSlices):
        table_name = "{0}_data_{1}".format(partition_prefix, i)
        query      = "DROP TABLE IF EXISTS {0}; CREATE TABLE {0} (row int, col int, rating float8);".format(table_name)
        output_vdbs.append( VictorDatabase(cur, [int4oid, int4oid, float8oid], table_name, 1) )
        
        cur.execute(query)

    if global_debug_flag:
        print "\t Creating Tables took {0}s".format(time.time() - t)
    
    # Partition the data
    row_bucket_size = max_rows/nSlices
    col_bucket_size = max_cols/nSlices

    t = time.time()
    # Any name for the cursor makes it server side, which is what we
    # want. Although we set arraysize here, pyscopg2 doesn't listen
    # (you can see it with the FETCH 1 messages so we have to remember
    # to add the buffer size we want with each fetchmany().
    cur            = conn.cursor('factor_parser_cursor_chicken') 
    cur.arraysize  = asize
    cur.execute( query_str )
    if global_debug_flag:
        print "\t Cursor execute completed {0}".format(time.time() - t)

    cdef int row, col
    cdef double rating
    cdef int new_row, new_col, row_bucket, col_bucket, zz
    
    if global_debug_flag:
        print "\t Cursor requesting {0} items per iteration".format(asize)
    
    results = cur.fetchmany(asize)
    while results:
        for (row,col,rating,) in results:
            new_row    = row_perm[ row - 1 ]
            new_col    = col_perm[ col - 1 ]
            row_bucket = get_bucket(new_row, row_bucket_size, nSlices)
            col_bucket = get_bucket(new_col, col_bucket_size, nSlices)
            zz         = row_bucket*nSlices + col_bucket
            if(nSlices == 1):
                zz = 0
            output_vdbs[zz].write_tuples([(row,col,rating)])
        results = cur.fetchmany(asize)
    
    # flush the remaining buffers
    for i in xrange(nSlices*nSlices):
        output_vdbs[i].flush()
        
    cur.close()
    conn.commit()
    conn.close()

    if global_debug_flag:
        print "\t Finished in {0}s".format(time.time() - t)

# TODO: Fix the partition model
cdef internal_partition_model(dbname, int seed, query_str, int nSlices, int epoch, int mid, int asize):
    victor_name    = "victor_{0}_{1}".format(mid,epoch)
    ssl_model_name = "{0}_lmodel".format(victor_name)
    ssr_model_name = "{0}_rmodel".format(victor_name)
    partition_prefix = victor_name

    vdb = get_VictorFile("{0}/{1}{2}.dat".format(dir_base, stats_simple_base, mid),
                         [int4oid, int4oid, int4oid],
                         mmap_wrapper.READ_MODE)
    (max_cols, max_rows, max_rank) = vdb.read_tuples()[0]

    row_perm = range(max_rows)
    col_perm = range(max_cols)
    # now do the permutation
    # the order is critical here!
    t = time.time()
    random.seed(seed)
    random.shuffle(row_perm)
    random.shuffle(col_perm)
    if global_debug_flag:
        print "\t [Partition Model] Finished Permutations in {0} for {1} rows and {2} cols".format(time.time() - t, max_rows, max_cols)
    

    # Partition the data
    row_bucket_size = max_rows/nSlices
    col_bucket_size = max_cols/nSlices

    [L,R,moviemean,stepsize,b,b2] = internal_retrieve_sparse_model_simple(dbname, mid)

    ## Build the L split tables and the output bucket
    output_vdbs = []
    for slice_no in xrange(nSlices):
        table_name = "{0}{1}".format(ssl_model_name, slice_no)

        output_vdbs.append ( get_VictorFile("{0}/{1}.dat".format(dir_base, table_name),
                                            [int4oid, float8arrayoid], 
                                            mmap_wrapper.WRITE_MODE) 
                             )


    for i in xrange(max_rows):
        ind = get_bucket(row_perm[i], row_bucket_size, nSlices)
        output_vdbs[ind].write_tuples([(i, L[i])])

    ###TODO: Do we need flushing here?
    for vdb in output_vdbs:
        vdb.flush()
    
    output_vdbs = []
    for slice_no in xrange(nSlices):
        table_name = "{0}{1}".format(ssr_model_name, slice_no)

        output_vdbs.append ( get_VictorFile("{0}/{1}.dat".format(dir_base, table_name), 
                                            [int4oid, float8arrayoid],
                                            mmap_wrapper.WRITE_MODE) 
                             )
    
    for j in xrange(max_cols):
        ind = get_bucket(col_perm[j], col_bucket_size, nSlices)
        output_vdbs[ind].write_tuples([(j, R[j])])
    
    ###TODO: Do we need flushing here?
    for vdb in output_vdbs:
        vdb.flush()

cdef internal_retrieve_sparse_model(dbname, int mid, int epoch_sign, int nsplits, int round, int chunk):
    import low_rank_helper

    vdb = get_VictorFile("{0}/{1}{2}.dat".format(dir_base, model_simple_base, mid), 
                         [int4oid, int4oid, int4oid, int4oid, float8oid, float8oid, float8oid], 
                         mmap_wrapper.READ_MODE)
    (nrows, l_rank, ncols, r_rank, moviemean, stepsize, b) = vdb.read_tuples()[0]    

    # Parse L
    l_name = low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))
    vdb = get_VictorFile("{0}/{1}.dat".format(dir_base, l_name), 
                         [int4oid, float8arrayoid], 
                         mmap_wrapper.READ_MODE)

    # NB: We want L to be sparse!
    # we use the fact that len(L[i]) == 0 tells us that we are not responsible for that row.
    L = [ [] for i in xrange(nrows)]

    for (i, l) in vdb.read_tuples():
        L[i] = l

    r_name = low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))
    vdb = get_VictorFile("{0}/{1}.dat".format(dir_base, r_name), 
                         [int4oid, float8arrayoid], 
                         mmap_wrapper.READ_MODE)

    R = [ [] for i in xrange(ncols)]
    
    for (i, r) in vdb.read_tuples():
        R[i] = r

    for_sd = (mid, epoch_sign, nsplits, round, chunk)
    return ([L,R,moviemean,stepsize,b,b*b], for_sd)


cdef internal_update_sparse_segment(dbname, int mid, int epoch_sign, int nsplits, int round, int chunk, L,R, double moviemean,double stepsize, double B, double B2, int asize):
    import low_rank_helper
    import psycopg2
    cdef int i

    l_name = low_rank_helper.tablename_lmodel(mid, epoch_sign, low_rank_helper.get_l_chunk(nsplits, round, chunk))
    
    vdb = get_VictorFile("{0}/{1}.dat".format(dir_base, l_name), 
                         [int4oid, float8arrayoid], 
                         mmap_wrapper.WRITE_MODE)

    for i in xrange(len(L)):
        if len(L[i]) > 0:
            vdb.write_tuples([(i, L[i])])

    vdb.flush()
            

    r_name = low_rank_helper.tablename_rmodel(mid, epoch_sign, low_rank_helper.get_r_chunk(nsplits, round, chunk))

    vdb = get_VictorFile("{0}/{1}.dat".format(dir_base, r_name), 
                         [int4oid, float8arrayoid], 
                         mmap_wrapper.WRITE_MODE)

    for i in xrange(len(R)):
        if len(R[i]) > 0:
            vdb.write_tuples([(i, R[i])])

    vdb.flush()


cdef internal_store_model(dbname, int mid, L, R, double moviemean, double stepsize, double B, int asize):
    table_name = "lr_model_instance"
    conn       = psycopg2.connect('dbname={0}'.format(dbname))
    cur        = conn.cursor()
    cur.execute("delete from {0} where mid={1}".format(table_name, mid))
    
    vdb = VictorDatabase(cur, [int4oid, float82darrayoid, float82darrayoid, float8oid, float8oid, float8oid], table_name, 1)
    vdb.write_tuples([(mid, L,R, moviemean, stepsize, B)])
    vdb.flush()
    
    cur.close()
    conn.commit()
    conn.close()

def internal_rebuild_model(dbname, int mid, int epoch_sign, int nsplits, int asize):
    from low_rank_helper import tablename_lmodel, tablename_rmodel
    cdef int i
    cdef double moviemean, stepsize, B
    
    table_name     = "lr_model_instance" 
    conn           = psycopg2.connect('dbname={0}'.format(dbname))
    cur            = conn.cursor()
    #TODO: write a function
    ##stats_query    = "SELECT array_upper(R,1) as max_cols, array_upper(L,1) as max_rows, array_upper(L,2) as max_rank, moviemean, stepsize, b from lr_model_instance where mid = {0};".format(mid)
    ##cur.execute(stats_query)
    if global_debug_flag:
        print "stats_query = {0}".format(stats_query)
    ##(num_cols, num_rows, max_rank, moviemean, stepsize, B) = cur.fetchone()
    vdb = get_VictorFile("{0}/{1}{2}.dat".format(dir_base, stats_base, mid),
                         [int4oid, int4oid, int4oid, float8oid, float8oid, float8oid],
                         mmap_wrapper.READ_MODE)
    (num_cols, num_rows, max_rank, moviemean, stepsize, B) = vdb.read_tuples()[0]
    
    L   = [[0.0]*max_rank for i in xrange(num_rows)]
    R   = [[0.0]*max_rank for i in xrange(num_cols)]
        
    for i in xrange(nsplits):
        vdb = get_VictorFile("{0}/{1}.dat".format(dir_base, tablename_lmodel(mid, epoch_sign, i)), 
                             [int4oid, float8arrayoid], 
                             mmap_wrapper.READ_MODE)
        if global_debug_flag:
            print "\t [L] copy_query = {0}".format(copy_query)

        for (i, x) in vdb.read_tuples():
            L[i] = x

    for i in xrange(nsplits):
        vdb = get_VictorFile("{0}/{1}.dat".format(dir_base, tablename_rmodel(mid, epoch_sign, i)), 
                             [int4oid, float8arrayoid], 
                             mmap_wrapper.READ_MODE)
        if global_debug_flag:
            print "\t [R] copy_query = {0}".format(copy_query)

        for (i, x) in vdb.read_tuples():
            R[i] = x

    cur.execute("delete from {0} where mid={1}".format(table_name, mid))
    vdb = VictorDatabase(cur, [int4oid, float82darrayoid, float82darrayoid, float8oid, float8oid, float8oid], table_name, 1)
    vdb.write_tuples([(mid, L, R, moviemean, stepsize, B)])
    vdb.flush()

    #Update dat files
    nRows = len(L)
    l_rank = len(L[0])
    nCols = len(R)
    r_rank = len(R[0])
    
    ## Write model_simple
    writeSingleTupleToFile("{0}/{1}{2}.dat".format(dir_base, model_simple_base, mid),
                           [int4oid, int4oid, int4oid, int4oid, float8oid, float8oid, float8oid],  
                           [(nRows, l_rank, nCols, r_rank, moviemean, stepsize, B)] )

    ## Write model_l_r
    writeSingleTupleToFile("{0}/{1}{2}.dat".format(dir_base, model_l_r_base, mid),
                           [float82darrayoid, float82darrayoid],  
                           [(L, R)] )
    ## Write stats_simple
    writeSingleTupleToFile("{0}/{1}{2}.dat".format(dir_base, stats_simple_base, mid),
                           [int4oid, int4oid, int4oid],
                           [(nCols, nRows, l_rank)] )      
    ## Write stats_simple
    writeSingleTupleToFile("{0}/{1}{2}.dat".format(dir_base, stats_base, mid),
                           [int4oid, int4oid, int4oid, float8oid, float8oid, float8oid],  
                           [(nCols, nRows, l_rank, moviemean, stepsize, B)] )

    cur.close()
    
    conn.commit()
    conn.close()

    
###########################################################################
# These are the front end for the permute and slice functions.
default_asize = 2000000
def permute_and_slice_data(dbname,seed, query_str, nSlices, epoch, mid,   asize=default_asize):
    internal_partition_data(dbname,seed, query_str, nSlices, epoch, mid, asize)

def permute_and_slice_model(dbname, seed, query_str, nSlices, epoch, mid, asize=default_asize):
    internal_partition_model(dbname, seed, query_str, nSlices, epoch, mid, asize)

## Sparse model functions.

def retrieve_sparse_model(dbname, mid, epoch_sign, nsplits, round, chunk):
    return internal_retrieve_sparse_model(dbname, mid, epoch_sign, nsplits, round, chunk)

def update_sparse_segment(dbname, mid, epoch_sign, nsplits, round, chunk, L, R, moviemean, stepsize, B, B2, asize=default_asize):
    internal_update_sparse_segment(dbname, mid, epoch_sign, nsplits, round, chunk, L, R, moviemean, stepsize, B, B2,asize)

## model store
def store_model(dbname, mid, L, R, moviemean, stepsize, B, B2, asize=default_asize):
    internal_store_model(dbname, mid, L, R, moviemean, stepsize, B, asize)

## rebuild the spread out model
def sparse_rebuild_model(dbname, mid, epoch_sign, nsplits, asize=default_asize):
    internal_rebuild_model(dbname, mid, epoch_sign, nsplits, asize)

def retrieve_sparse_model_simple(dbname, mid):
    return internal_retrieve_sparse_model_simple(dbname,mid)
