"""
This module contains TupleSchema, TupleReader and TupleWriter classes.
"""

from src.vsql_core.utils.factored_parser.victor_buffer import *
from src.vsql_core.utils.factored_parser.atomic_data_processor import *
import struct
import cStringIO
from src.vsql_core.utils.factored_parser.data_types import *

############################################################
### Tuple Processor Module
### 
### This module contains TupleSchema, TupleReader and 
### TupleWriter classes
### TupleReader takes tupleschema and returns tuple
### TupleWriter takes tupleschema and writes the tuple
############################################################

#Read functions for oids
oid_type_read_fncs = [Atomic_Data_Processor().read_char,
                      Atomic_Data_Processor().read_bool,                      
                      Atomic_Data_Processor().read_int2,
                      Atomic_Data_Processor().read_int4,
                      Atomic_Data_Processor().read_int8,
                      Atomic_Data_Processor().read_float4,
                      Atomic_Data_Processor().read_float8,
                      Atomic_Data_Processor().read_char_array,
                      Atomic_Data_Processor().read_bool_array,
                      Atomic_Data_Processor().read_int2_array,
                      Atomic_Data_Processor().read_int4_array,
                      Atomic_Data_Processor().read_int8_array,
                      Atomic_Data_Processor().read_float4_array,
                      Atomic_Data_Processor().read_float8_array,
                      Atomic_Data_Processor().read_float8_2d_array,
                      Atomic_Data_Processor().read_partition_tuple
                      ]

#Write functions for oids
oid_type_write_fncs = [Atomic_Data_Processor().write_char,
                       Atomic_Data_Processor().write_bool,                      
                       Atomic_Data_Processor().write_int2,
                       Atomic_Data_Processor().write_int4,
                       Atomic_Data_Processor().write_int8,
                       Atomic_Data_Processor().write_float4,
                       Atomic_Data_Processor().write_float8,
                       Atomic_Data_Processor().write_char_array,
                       Atomic_Data_Processor().write_bool_array,
                       Atomic_Data_Processor().write_int2_array,
                       Atomic_Data_Processor().write_int4_array,
                       Atomic_Data_Processor().write_int8_array,
                       Atomic_Data_Processor().write_float4_array,
                       Atomic_Data_Processor().write_float8_array,
                       Atomic_Data_Processor().write_float8_2d_array,
                       Atomic_Data_Processor().write_partition_tuple
                       ]
# Oid_dictionary: key oid
oid_dict = dict()
i = 0
for  (oid,moniker,struct_type,python_type,size) in oid_table:
    oid_dict[oid] = (moniker,struct_type,python_type,size, oid_type_read_fncs[i], oid_type_write_fncs[i])
    i = i + 1

"""
TupleSchema class takes oid list and
resolves types: produces information about the oid
"""
class TupleSchema:
    def __init__(self, oid_list):
        self.schema = oid_list

    """
    From oid list, returns information for the data types in the schema
    @rype: array
    @return: information for the data types in the schema
    """
    def get_resolved_types(self):
        resolved_types = []

        for elem_type in self.schema:
            resolved_types.append(oid_dict[elem_type])
            
        return resolved_types
        
psql_header_read = bytearray(b'PGCOPY\n') + bytearray(b'\377\r\n\0')
psql_header_write = b'PGCOPY\n\377\r\n\0\0\0\0\0\0\0\0\0'

"""
TupleReader takes tupleschema and returns tuple
"""
class TupleReader:
    def __init__(self, b, ts, check_header = False):
        self.function_list = [(size, read_fnc) 
                              for (moniker,struct_type,python_type,size, read_fnc, write_fnc) in ts.get_resolved_types()]
        self.buf = VictorBuffer(b)
        if check_header:
            if len(b) <= 11 or not (b[0:11] == psql_header_read):
                print "lenb={1} b={0}".format(b,len(b))
                assert(False)
                
            self.buf.advance_ci(11) # skip the header
            flags_field   = Atomic_Data_Processor().read_int4_without_length(self.buf)
            header_extension_length = Atomic_Data_Processor().read_int4_without_length(self.buf)
            self.buf.advance_ci(header_extension_length) # skip over the remaining 0 part of the header

    def read_tuple_for_partition(self):
        if(self.buf.process_more()):
            # we assume that a buffer contains whole tuple
            ret = []
            (l, f) = self.function_list[0]
            ret_t = f(self.buf)
            return ret_t
        else:
            return None

    """
    Reads tuple from the given buffer and returns it
    @rtype: tuple
    @return: tuple in the form of the given schema
    """
    def read_tuple(self):
        #Read n fields
        nFields = Atomic_Data_Processor().read_int2_without_length(self.buf)
        if (nFields > 0):
            #Check its consistent with tuple size
            assert(nFields == len(self.function_list))
        
            if(self.buf.process_more()):
                # we assume that a buffer contains whole tuple
                ret = []
                for (l, f) in self.function_list:
                    ret.append( f(self.buf) )

                return ret
            else:
                return None
    
"""
TupleWriter takes tupleschema and writes the tuple
"""
class TupleWriter:
    def __init__(self, cur, table_name, ts, check_header = False, nBuf = 10000):
        self.cur           = cur
        self.query         = "COPY {table_name} FROM STDIN WITH BINARY;".format(table_name = table_name) 
        self.nBuf          = nBuf
        self.function_list = [(size, write_fnc) 
                              for (moniker,struct_type,python_type,size, read_fnc, write_fnc) in ts.get_resolved_types()]
        self.ch = check_header
        self.init_buffer()
        

    def init_buffer(self):
        self.buf = cStringIO.StringIO()
        if self.ch:
            self.buf.write(psql_header_write)
        self.index = 0

    def flush(self):
        self.write_eos()
        tmp = cStringIO.StringIO(self.buf.getvalue())
        self.cur.copy_expert(self.query, tmp)
        tmp.close()

    # flush, then free all the resources
    def close(self):
        self.flush()
        self.buf.close()

    def check_flush(self):
        if self.index > self.nBuf:
            self.flush()
            self.buf.close()
            self.init_buffer()

    def write_tuple_for_partition(self, tup):
        self.buf.write(struct_int2.pack(tup.num_field)) #write nFields
        (l, f) = self.function_list[0]
        f(self.buf, tup.tuple_data)

    """
    Writes tuple to the given buffer
    @type tup: tuple
    @param tup: tuple in the form of the given schema to write to the buffer
    """
    def write_tuple(self, tup):
        i = 0

        self.buf.write(struct_int2.pack(len(self.function_list))) #write nFields
        self.index += 2
        for (l, f) in self.function_list:
            self.index += f(self.buf, tup[i])
            i = i + 1

        self.check_flush()
    
    def write_eos(self):
        self.buf.write( struct_int2.pack( -1 ) ) 
