---------------------------------------
-- Permutation and ordering fucntions -
---------------------------------------
DROP FUNCTION IF EXISTS create_permutations(int, int, int) CASCADE; 
CREATE FUNCTION create_permutations(nrows int, ncols int, seed int)
RETURNS VOID
AS $$
import random
random.seed(seed)
row_map = range(nrows)
col_map = range(ncols)
random.shuffle(row_map)
random.shuffle(col_map)
GD['row_map'] = row_map
GD['col_map'] = col_map
$$ LANGUAGE plpythonu;

DROP FUNCTION IF EXISTS apply_permutation(int) CASCADE; 
CREATE FUNCTION apply_permutation(i int)
RETURNS int
AS $$
the_map = GD['permutation']
return the_map[i]
$$ LANGUAGE plpythonu STABLE;


-------------------------------------------------
--- Chunk and round are used by VSQL
--- to select the appropriate part of the matrix
--------------------------------------------------
DROP FUNCTION IF EXISTS round(int, int, int, int, int) CASCADE; 
CREATE FUNCTION round(nsplits int, nrows int, ncols int, row_ii int, col_ii int)
RETURNS INT
AS $$
	row_map = GD['row_map']
	row_i   = row_map[row_ii]
	col_map = GD['col_map']
	col   = col_map[col_ii]

	row_bucket_size = nrows/nsplits
	col_bucket_size = ncols/nsplits
	row_bucket = row_i/row_bucket_size
	col_bucket = col/col_bucket_size
	return (col_bucket - row_bucket) % nsplits
$$ LANGUAGE plpythonu STABLE;

DROP FUNCTION IF EXISTS chunk(int, int, int, int, int) CASCADE; 
CREATE FUNCTION chunk(nsplits int, nrows int, ncols int, row_ii int, col_ii int)
RETURNS INT
AS $$
	row_map = GD['row_map']
	row_i   = row_map[row_ii]
	col_map = GD['col_map']
	col   = col_map[col_ii]

	row_bucket_size = nrows/nsplits
	col_bucket_size = ncols/nsplits
	row_bucket = row_i/row_bucket_size
	col_bucket = col/col_bucket_size
	return col_bucket
$$ LANGUAGE plpythonu STABLE;

-----------------------------------------------------
-- Accessors for the model. Intended for debugging --
-----------------------------------------------------
DROP FUNCTION IF EXISTS get_l_model(i int) CASCADE; 
CREATE FUNCTION get_l_model(i int) RETURNS float8[]
AS $$
   [L,R,moviemean, stepsize,b,b2] = GD['lr_model']
   return L[i]
$$ LANGUAGE plpythonu STABLE;

DROP FUNCTION IF EXISTS get_r_model(i int) CASCADE; 
CREATE FUNCTION get_r_model(i int) RETURNS float8[]
AS $$
   [L,R,moviemean, stepsize,b,b2] = GD['lr_model']
   return R[i]
$$ LANGUAGE plpythonu STABLE;

-- We need to be able to retreive it from the databse.

-- these can be used as
-- UPDATE model SET model[i] = get_l_model(i)
-- where modelid = 0;


---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
-- This code is used at the end of a (parallel) update in V-SQL.
-- It should be executed transactionally (else updates may be lost)
-- It reads in the model, updates its fragment in memory and writes it back to disk.
-- The fragment is computed from (chunk,round)
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
DROP FUNCTION IF EXISTS update_fragment(int, int, int, int, int,int,int) CASCADE; 
CREATE FUNCTION update_fragment(max_rank int, nsplits int, nrows int, ncols int, chunk int, round int, mid int)
RETURNS VOID
AS $$   
   
   row_bucket_size = nrows/nsplits	
   col_bucket_size = ncols/nsplits	
   
   col_bucket      = chunk
   col_low         = chunk*col_bucket_size
   col_high        = min((chunk+1)*col_bucket_size, ncols)
   
   row_bucket      = (- (round - col_bucket)) % nsplits   
   row_low         = row_bucket*row_bucket_size
   row_high        = min((row_bucket+1)*row_bucket_size, nrows) 
   
   [L_stored,R_stored, mean, stepsize, b, z] = GD['lr_model']	

   ## HERE We read in the models   
   v_plan = plpy.prepare("SELECT array_upper(L,1) as nRows, array_upper(L,2) as l_rank, array_upper(R,1) as ncols, array_upper(R,2) as r_rank FROM lr_model_instance WHERE mid=$1", ["int"])
   l_plan = plpy.prepare("SELECT array_to_string(L,',') as l FROM lr_model_instance where mid = $1", ["int"]);
   r_plan = plpy.prepare("SELECT array_to_string(R,',') as r FROM lr_model_instance where mid = $1", ["int"]);
   
   rv = plpy.execute(v_plan, [mid], 1)
   nRows     = rv[0]["nrows"]
   l_rank    = rv[0]["l_rank"]
   nCols     = rv[0]["ncols"]
   r_rank    = rv[0]["r_rank"]

   # build L and R
   L = []
   l_str = plpy.execute(l_plan, [mid],1)[0]["l"]
   ls    = [float(x) for x in l_str.split(',')]
   for i in range(nRows):
       li = []	
       for k in range(l_rank):
           li.append(ls[i * l_rank + k])
       L.append(li)

   R = []
   r_str = plpy.execute(r_plan, [mid],1)[0]["r"]
   rs    = [float(x) for x in l_str.split(',')]
   for i in range(nCols):
       ri = []	
       for k in range(r_rank):
       	   ri.append(rs[i * r_rank + k])
       R.append(ri)
   plpy.info("Finished building L and R")
   
   # Overwrite our chunk
   for i in range(row_low, row_high):	  	  
     L[i] = L_stored[i]

   for j in range(col_low, col_high):
     R[j] = R_stored[j]

   plpy.info("Finished Copying fragment")

   # build the string for insertion
   L_str = []
   for l in L:
       L_str.append("{" + ','.join([str(y) for y in l]) + "}")
       
   string_L = "'{" + ','.join(L_str) + "}'" 

   R_str = []
   for r in R:
       R_str.append('{' + ','.join([str(y) for y in r]) + '}')

   string_R = "'{" + ','.join(R_str) + "}'"
   plpy.execute("UPDATE lr_model_instance SET L="+ string_L + ", R=" + string_R + " WHERE mid=" + str(mid))
$$ LANGUAGE plpythonu;



---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
-- This code is used at the end of a (parallel) update in V-SQL.
-- It should be executed transactionally (else updates may be lost)
-- It reads in the model, updates its fragment in memory and writes it back to disk.
-- The fragment is computed from (chunk,round)
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
DROP FUNCTION IF EXISTS overwrite_fragment(int, int, int, int, int,int,int) CASCADE; 
CREATE FUNCTION overwrite_fragment(max_rank int, nsplits int, nrows int, ncols int, chunk int, round int, mid int)
RETURNS VOID
AS $$   

   row_bucket_size = nrows/nsplits	
   col_bucket_size = ncols/nsplits	
 
   assert(row_bucket_size > 0 and col_bucket_size > 0)   

   col_bucket      = chunk
   col_low         = chunk*col_bucket_size
   col_high        = min((chunk+1)*col_bucket_size, ncols) - 1
   
   row_bucket      = (- (round - col_bucket)) % nsplits   
   row_low         = row_bucket*row_bucket_size
   row_high        = min((row_bucket+1)*row_bucket_size, nrows) - 1
   
   [L_stored,R_stored, mean, stepsize, b, z] = GD['lr_model']	


   # build strings for insert
   L_str = "'{" + ','.join(["{" + ','.join([str(y) for y in l]) + "}" for l in L_stored[row_low:(row_high+1)]]) + "}'"
   R_str = "'{" + ','.join(["{" + ','.join([str(y) for y in r]) + "}" for r in R_stored[col_low:(col_high+1)]]) + "}'"

   # +1 to compensate for postgresql indexing   
   plpy.execute("UPDATE lr_model_instance SET L[%d:%d][1:%d]=%s, R[%d:%d][1:%d]=%s WHERE mid=%d" % (row_low + 1, row_high + 1, max_rank, L_str, col_low + 1, col_high + 1, max_rank, R_str,mid))
$$ LANGUAGE plpythonu;






DROP FUNCTION IF EXISTS store_model(int) CASCADE; 
CREATE FUNCTION store_model(mid int)
RETURNS VOID
AS $$   
   [L_stored,R_stored, mean, stepsize, b, z] = GD['lr_model']	   
   L_str = []
   for l in L_stored:
       L_str.append("{" + ','.join([str(y) for y in l]) + "}")       
   string_L = "'{" + ','.join(L_str) + "}'" 

   R_str = []
   for r in R_stored:
       R_str.append("{" + ','.join([str(y) for y in r]) + "}")

   string_R = "'{" + ','.join(R_str) + "}'"
   #plan = plpy.prepare("INSERT INTO lr_model_instance VALUES ($1,$2,$3,$4,$5,$6)", ["int", "float8[][]", "float8[][]", "float8", "float8", "float8"])
   #plpy.execute(plan, [mid, string_L, string_R, mean, stepsize, b])
   the_values = [str(mid), string_L, string_R, str(mean), str(stepsize), str(b)]
   exec_str   = "INSERT INTO lr_model_instance VALUES (" + ','.join(the_values) + ")"
   plpy.execute(exec_str)	   
$$ LANGUAGE plpythonu;


--------------------------
--- Model Initialization
--------------------------
DROP FUNCTION IF EXISTS initialize_model(int, int, int, float8, float8, float8) CASCADE; 
CREATE FUNCTION initialize_model(max_rank int, nrows int, ncols int, b float8, mean float8, stepsize float8)
RETURNS VOID
AS $$
   import low_rank_helper
   (L,R) = low_rank_helper.make_initial(max_rank, nrows, ncols, b)
   GD['lr_model'] = [L,R, mean, stepsize, b, b*b]
$$ LANGUAGE plpythonu;


--------------------------------------------------------------
-- Given a model id, this fetches it and [puts it into memory
-- TODO: Prepare the plans across invocations
DROP FUNCTION IF EXISTS retrieve_model_instance(int) CASCADE; 
CREATE FUNCTION retrieve_model_instance(mid int)
RETURNS VOID
AS $$
   v_plan = plpy.prepare("SELECT array_upper(L,1) as nRows, array_upper(L,2) as l_rank, array_upper(R,1) as ncols, array_upper(R,2) as r_rank, moviemean as moviemean, stepsize as stepsize, b as b FROM lr_model_instance WHERE mid=$1", ["int"])
   l_plan = plpy.prepare("SELECT L[$1] as l FROM lr_model_instance where mid = $2", ["int", "int"]);
   r_plan = plpy.prepare("SELECT R[$1] as r FROM lr_model_instance where mid = $2", ["int", "int"]);
   
   rv = plpy.execute(v_plan, [mid], 1)
   nRows     = rv[0]["nrows"]
   l_rank    = rv[0]["l_rank"]
   nCols     = rv[0]["ncols"]
   r_rank    = rv[0]["r_rank"]
   stepsize  = rv[0]["stepsize"]
   moviemean = rv[0]["moviemean"]
   b         =  rv[0]["b"]

   L = []
   for i in range(1,nRows+1):
          rv = plpy.execute(l_plan, [i,mid])	
          L.append(rv[0]["l"]);
   R = []
   for i in range(1,nCols+1):
          rv = plpy.execute(r_plan, [i,mid])	
          R.append(rv[0]["r"]);

   GD['lr_model'] = [L,R,moviemean,stepsize, b, b*b]
$$ LANGUAGE plpythonu ;


--------------------------------------------------------------
-- Given a model id, this fetches it and [puts it into memory
-- TODO: Prepare the plans across invocations
DROP FUNCTION IF EXISTS retrieve_model_instance_string(int) CASCADE; 
CREATE FUNCTION retrieve_model_instance_string(mid int)
RETURNS VOID
AS $$
   v_plan = plpy.prepare("SELECT array_upper(L,1) as nRows, array_upper(L,2) as l_rank, array_upper(R,1) as ncols, array_upper(R,2) as r_rank, moviemean as moviemean, stepsize as stepsize, b as b FROM lr_model_instance WHERE mid=$1", ["int"])
   l_plan = plpy.prepare("SELECT array_to_string(L,',') as l FROM lr_model_instance where mid = $1", ["int"]);
   r_plan = plpy.prepare("SELECT array_to_string(R,',') as r FROM lr_model_instance where mid = $1", ["int"]);
   
   rv = plpy.execute(v_plan, [mid], 1)
   nRows     = rv[0]["nrows"]
   l_rank    = rv[0]["l_rank"]
   nCols     = rv[0]["ncols"]
   r_rank    = rv[0]["r_rank"]
   stepsize  = rv[0]["stepsize"]
   moviemean = rv[0]["moviemean"]
   b         =  rv[0]["b"]

   L = []
   l_str = plpy.execute(l_plan, [mid],1)[0]["l"]
   ls    = [float(x) for x in l_str.split(',')]
   for i in range(nRows):
       li = []	
       for k in range(l_rank):
           li.append(ls[i * l_rank + k])
       L.append(li)

   R = []
   r_str = plpy.execute(r_plan, [mid],1)[0]["r"]
   rs    = [float(x) for x in r_str.split(',')]
   for i in range(nCols):
       ri = []	
       for k in range(r_rank):
       	   ri.append(rs[i * r_rank + k])
       R.append(ri)

   GD['lr_model'] = [L,R,moviemean,stepsize, b, b*b]
$$ LANGUAGE plpythonu;


-- This fucntion takes a step on a particular data item
-- It assumes that the model is stored in lr_model
DROP FUNCTION IF EXISTS take_step(int, int, float8) CASCADE; 
CREATE FUNCTION take_step(i int, j int, rating float8)
RETURNS VOID
AS $$
   import low_rank_helper
   [L,R,moviemean,stepsize,B,B2] = GD['lr_model']
   err = low_rank_helper.dot(L[i],R[j]) + moviemean - rating	
   e   = -(stepsize * err)
   tempLi = list(L[i])
   low_rank_helper.scale_and_add(tempLi, R[j],e)   			    
   low_rank_helper.scale_and_add(R[j],   L[i],e)
   L[i] = tempLi
   low_rank_helper.ball_project(L[i],B,B2)
   low_rank_helper.ball_project(R[j],B,B2)
   GD['lr_model'] =    [L,R,moviemean,stepsize,B,B2] 
$$ LANGUAGE plpythonu;




-- This fucntion takes a step on a particular data item
-- It assumes that the model is stored in lr_model
DROP FUNCTION IF EXISTS model_print() CASCADE; 
CREATE FUNCTION model_print()
RETURNS VOID
AS $$
   [L,R,moviemean,stepsize,B,B2] = GD['lr_model']
   plpy.info(str(moviemean) + " " + str(stepsize) + " " + str(B) + " " + str(B2) + " l =" + str(L[0]))
$$ LANGUAGE plpythonu;




--------------------------
-- Compute the objective
-- mean-squared error
DROP FUNCTION IF EXISTS mse(int, int, float8) CASCADE; 
CREATE FUNCTION mse(i int, j int, rating float8)
RETURNS float8
AS $$
   import low_rank_helper
   [L,R,moviemean,stepsize,B,B2] = GD['lr_model']
   err = low_rank_helper.dot(L[i],R[j]) + moviemean - rating	
   return err*err
$$ LANGUAGE plpythonu;

-- Compute the objective
-- select SQRT(SUM(mse(row - 1, col - 1, rating)))/SQRT(COUNT(*)) from probe_table;


