"""
This module implements the unit tests for
read and write of the atomic data types by TupleReader and TupleWriter.
"""

import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from unittests.unittest_helper import *
from src.vsql_core.utils.factored_parser.tuple_processor import *
import cStringIO

class TupleRoundTripRegressionTests(unittest.TestCase):
    
    def generic_test_round_trip_data(self, schema, type_params, method_name):
        """
        Tests whether random data types are successfully read and written by TupleReader and TupleWriter
        @type schema: tuple
        @param schema: represents the data types in the tuple
        @type type_params: tuple
        @param type_params: contains information(is_array_type, elm_size, random_fnc, elm_type, pack_fnc)
        about each data type in the tuple.
        @type method_name: string
        @param method_name: the name of the method who calls this generic method
        """
        for i in range(test_num):
            b = cStringIO.StringIO()

            rand_tuple = []
            for (is_array_type, elm_size, random_fnc, elm_type, pack_fnc) in type_params:
                # create data
                data = create_rand_data(is_array_type, random_fnc)

                #append to the tuple
                rand_tuple.append(data)

            ts = TupleSchema(schema)
            tw = TupleWriter(b, ts)
            tw.write_tuple(rand_tuple)
            tr = TupleReader(b.getvalue(), ts)
            ret_tuple = tr.read_tuple()

            compare_tuples(self, type_params, rand_tuple, ret_tuple, method_name, exact_equal)

    
    def test_round_trip_single_char_tuple(self):
        """
        Tests whether tuples that only contain random char are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([charoid], [char_type_param], 'round_trip_single_char_tuple')

    
    def test_round_trip_single_bool_tuple(self):
        """
        Tests whether tuples that only contain random bool are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([booloid], [bool_type_param], 'round_trip_single_bool_tuple')

    
    def test_round_trip_single_int2_tuple(self):
        """
        Tests whether tuples that only contain random int2 are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([int2oid], [int2_type_param], 'round_trip_single_int2_tuple')

    
    def test_round_trip_single_int4_tuple(self):
        """
        Tests whether tuples that only contain random int4 are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([int4oid], [int4_type_param], 'round_trip_single_int4_tuple')

    
    def test_round_trip_single_int8_tuple(self):
        """
        Tests whether tuples that only contain random int8 are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([int8oid], [int8_type_param], 'round_trip_single_int8_tuple')

    
    def test_round_trip_single_float4_tuple(self):
        """
        Tests whether tuples that only contain random float4 are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([float4oid], [float4_type_param], 'round_trip_single_float4_tuple')

    
    def test_round_trip_single_float8_tuple(self):
        """
        Tests whether tuples that only contain random float8 are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([float8oid], [float8_type_param], 'round_trip_single_float8_tuple')

    
    def test_round_trip_single_char_array_tuple(self):
        """
        Tests whether tuples that only contain random char array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([chararrayoid], [char_array_type_param], 'round_trip_single_char_array_tuple')

    
    def test_round_trip_single_bool_array_tuple(self):
        """
        Tests whether tuples that only contain random bool array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([boolarrayoid], [bool_array_type_param], 'round_trip_single_bool_array_tuple')

    
    def test_round_trip_single_int2_array_tuple(self):
        """
        Tests whether tuples that only contain random int2 array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([int2arrayoid], [int2_array_type_param], 'round_trip_single_int2_array_tuple')

    
    def test_round_trip_single_int4_array_tuple(self):
        """
        Tests whether tuples that only contain random int4 array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([int4arrayoid], [int4_array_type_param], 'round_trip_single_int4_array_tuple')

    
    def test_round_trip_single_int8_array_tuple(self):
        """
        Tests whether tuples that only contain random int8 array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([int8arrayoid], [int8_array_type_param], 'round_trip_single_int8_array_tuple')

    
    def test_round_trip_single_float4_array_tuple(self):
        """
        Tests whether tuples that only contain random float4 array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([float4arrayoid], [float4_array_type_param], 'round_trip_single_float4_array_tuple')

    
    def test_round_trip_single_float8_array_tuple(self):
        """
        Tests whether tuples that only contain random float8 array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([float8arrayoid], [float8_array_type_param], 'round_trip_single_float8_array_tuple')

    
    def test_round_trip_single_float8_2d_array_tuple(self):
        """
        Tests whether tuples that only contain random float8 2d array are successfully read and written 
        by TupleReader and TupleWriter
        """
        self.generic_test_round_trip_data([float82darrayoid], [float8_2d_array_type_param], 
                                          'round_trip_single_float8_2d_array_tuple')

    
    def test_tuples_with_non_array_data(self):
        """
        This method tests read and write of tuples with possible non-array data type combinations.
        Since there are 7 non-array data types, all combinations produce 127 different schema.
        """
        test_tuples_with_non_array_data(self.generic_test_round_trip_data, limit_test, without_sql)

    
    def test_tuples_with_array_data(self):
        """
        This method tests read and write of tuples with possible array(1d or 2d) data type combinations.
        Since there are 8 array data types, all combinations produce 255 different schema.
        """
        test_tuples_with_array_data(self.generic_test_round_trip_data, limit_test, without_sql)

    
    def test_tuples_with_data(self):
        """
        This method tests read and write of tuples with possible non-array and array(1d or 2d) data type combinations.
        Since there are 15 data types, all combinations should produce 2 ** 15 - 1 different schema.
        """
        test_tuples_with_data(self.generic_test_round_trip_data, limit_test, without_sql)
