"""
This module implements the unit tests for
read of the atomic data types by VictorDatabase.
"""

import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from unittests.unittest_helper import *
from src.vsql_core.utils.factored_parser.tuple_processor import *
from src.vsql_core.utils.factored_parser.psql_util import *
import cStringIO


class VictorDBReadRegressionTests(unittest.TestCase):
    
    def generic_test_read_data(self, schema, type_params, method_name):
        """
        Tests whether random data types are successfully read by VictorDatabase
        @type schema: tuple
        @param schema: represents the data types in the tuple
        @type type_params: tuple
        @param type_params: contains information(is_array_type, elm_size, random_fnc, elm_type, pack_fnc, convert_sql_fnc, 
        sql_create_type) about each data type in the tuple.
        @type method_name: string
        @param method_name: the name of the method who calls this generic method
        """
        db_conn = psycopg2.connect('dbname={0}'.format(dbname))
        cur = db_conn.cursor()

        for i in range(test_num):
            create_table_query = create_table_from_fields(type_params)
            cur.execute(create_table_query)

            all_tuples = []
            for j in range(tuple_size):
                rand_tuple = []
                converted_data = []
                insert_query = 'INSERT INTO ' + temp_table_name + ' VALUES ('
                for (is_array_type, elm_size, random_fnc, elm_type, pack_fnc, convert_sql_fnc, sql_create_type) in type_params:
                    # create rand data
                    data = create_rand_data(is_array_type, random_fnc)

                    # append to the tuple
                    rand_tuple.append(data)
                    converted_data.append(convert_sql_fnc(data))


                insert_query += ','.join(converted_data)
                insert_query += ');'
                cur.execute(insert_query)
                all_tuples.append(rand_tuple)

            db_conn.commit()

            vdb = VictorDatabase(cur, schema, temp_table_name, 1)
            ret_tuples = vdb.read_tuples()

            compare_set_of_tuples(self, type_params, all_tuples, ret_tuples, method_name, exact_equal)

    
    def test_read_single_char_type(self):
        """
        Tests whether tuples that only contain random char are successfully read by VictorDatabase
        """
        self.generic_test_read_data([charoid], [char_vdb_type_param], 'read_single_char_type')

    
    def test_read_single_bool_type(self):
        """
        Tests whether tuples that only contain random bool are successfully read by VictorDatabase
        """
        self.generic_test_read_data([booloid], [bool_vdb_type_param], 'read_single_bool_type')

    
    def test_read_single_int2_type(self):
        """
        Tests whether tuples that only contain random int2 are successfully read by VictorDatabase
        """
        self.generic_test_read_data([int2oid], [int2_vdb_type_param], 'read_single_int2_type')

    
    def test_read_single_int4_type(self):
        """
        Tests whether tuples that only contain random int4 are successfully read by VictorDatabase
        """
        self.generic_test_read_data([int4oid], [int4_vdb_type_param], 'read_single_int4_type')

    
    def test_read_single_int8_type(self):
        """
        Tests whether tuples that only contain random int8 are successfully read by VictorDatabase
        """
        self.generic_test_read_data([int8oid], [int8_vdb_type_param], 'read_single_int8_type')

    
    def test_read_single_float4_type(self):
        """
        Tests whether tuples that only contain random float4 are successfully read by VictorDatabase
        """
        self.generic_test_read_data([float4oid], [float4_vdb_type_param], 'read_single_float4_type')

    
    def test_read_single_float8_type(self):
        """
        Tests whether tuples that only contain random float8 are successfully read by VictorDatabase
        """
        self.generic_test_read_data([float8oid], [float8_vdb_type_param], 'read_single_float8_type')

    
    def test_read_single_bool_array_type(self):
        """
        Tests whether tuples that only contain random bool array are successfully read by VictorDatabase
        """
        self.generic_test_read_data([boolarrayoid], [bool_array_vdb_type_param], 'read_single_bool_array_type')

    
    def test_read_single_int2_array_type(self):
        """
        Tests whether tuples that only contain random int2 array are successfully reand by VictorDatabase
        """
        self.generic_test_read_data([int2arrayoid], [int2_array_vdb_type_param], 'read_single_int2_array_type')

    
    def test_read_single_int4_array_type(self):
        """
        Tests whether tuples that only contain random int4 array are successfully read by VictorDatabase
        """
        self.generic_test_read_data([int4arrayoid], [int4_array_vdb_type_param], 'read_single_int4_array_type')

    
    def test_read_single_int8_array_type(self):
        """
        Tests whether tuples that only contain random int8 array are successfully read by VictorDatabase
        """
        self.generic_test_read_data([int8arrayoid], [int8_array_vdb_type_param], 'read_single_int8_array_type')

    
    def test_read_single_float4_array_type(self):
        """
        Tests whether tuples that only contain random float4 array are successfully read by VictorDatabase
        """
        self.generic_test_read_data([float4arrayoid], [float4_array_vdb_type_param], 'read_single_float4_array_type')

    
    def test_read_single_float8_array_type(self):
        """
        Tests whether tuples that only contain random float8 array are successfully read by VictorDatabase
        """
        self.generic_test_read_data([float8arrayoid], [float8_array_vdb_type_param], 'read_single_float8_array_type')

    
    def test_read_single_float8_2d_array_type(self):
        """
        Tests whether tuples that only contain random float8 2d array are successfully read by VictorDatabase
        """
        self.generic_test_read_data([float82darrayoid], [float8_2d_array_vdb_type_param], 'read_single_float8_2d_array_type')

    
    def test_tuples_with_non_array_data(self):
        """
        This method tests read of tuples with possible non-array data type combinations.
        Since there are 7 non-array data types, all combinations produce 127 different schema.
        """
        test_tuples_with_non_array_data(self.generic_test_read_data, limit_test, with_sql)

    
    def test_tuples_with_array_data(self):
        """
        This method tests read of tuples with possible array(1d or 2d) data type combinations.
        Since there are 7 array data types, all combinations produce 127 different schema.
        """
        test_tuples_with_array_data(self.generic_test_read_data, limit_test, with_sql)

    
    def test_tuples_with_data(self):
        """
        This method tests read of tuples with possible non-array and array(1d or 2d) data type combinations.
        Since there are 14 data types, all combinations should produce 2 ** 14 - 1 different schema.
        """
        test_tuples_with_data(self.generic_test_read_data, limit_test, with_sql)





