"""
This module implements the unit tests for
write of the atomic data types by VictorDatabase.
"""

import random
import unittest
import StringIO
import HTMLTestRunner
import traceback
import sys

from unittests.unittest_helper import *
from src.vsql_core.utils.factored_parser.tuple_processor import *
from src.vsql_core.utils.factored_parser.psql_util import *
import cStringIO

class VictorDBWriteRegressionTests(unittest.TestCase):
    
    def generic_test_write_data(self, schema, type_params, method_name):
        """
        Tests whether random data types are successfully written by VictorDatabase
        @type schema: tuple
        @param schema: represents the data types in the tuple
        @type type_params: tuple
        @param type_params: contains information(is_array_type, elm_size, random_fnc, elm_type, pack_fnc, convert_sql_fnc, 
        sql_create_type) about each data type in the tuple.
        @type method_name: string
        @param method_name: the name of the method who calls this generic method
        """
        db_conn = psycopg2.connect('dbname={0}'.format(dbname))
        cur = db_conn.cursor()

        for i in range(test_num):
            create_table_query = create_table_from_fields(type_params)
            cur.execute(create_table_query)

            all_tuples = []
            for j in range(tuple_size):
                rand_tuple = []

                for (is_array_type, elm_size, random_fnc, elm_type, pack_fnc, convert_sql_fnc, sql_create_type) in type_params:
                    # create rand data
                    data = create_rand_data(is_array_type, random_fnc)

                    # append to the tuple
                    rand_tuple.append(data)

                all_tuples.append(rand_tuple)

            db_conn.commit()

            vdb = VictorDatabase(cur, schema, temp_table_name, 1)
            vdb.write_tuples(all_tuples)
            vdb.flush()

            # not read the db and compare
            cur.execute('SELECT * FROM ' + temp_table_name + ';')
            ret_tuples = []
            for record in cur:
                ret_tuples.append(record)
                
            compare_set_of_tuples(self, type_params, all_tuples, ret_tuples, method_name, almost_equal)

    
    def test_write_single_char_type(self):
        """
        Tests whether tuples that only contain random char are successfully written by VictorDatabase
        """
        self.generic_test_write_data([charoid], [char_vdb_type_param], 'write_single_char_type')

    
    def test_write_single_bool_type(self):
        """
        Tests whether tuples that only contain random bool are successfully written by VictorDatabase
        """
        self.generic_test_write_data([booloid], [bool_vdb_type_param], 'write_single_bool_type')

    
    def test_write_single_int2_type(self):
        """
        Tests whether tuples that only contain random int2 are successfully written by VictorDatabase
        """
        self.generic_test_write_data([int2oid], [int2_vdb_type_param], 'write_single_int2_type')

    
    def test_write_single_int4_type(self):
        """
        Tests whether tuples that only contain random int4 are successfully written by VictorDatabase
        """
        self.generic_test_write_data([int4oid], [int4_vdb_type_param], 'write_single_int4_type')

    
    def test_write_single_int8_type(self):
        """
        Tests whether tuples that only contain random int8 are successfully written by VictorDatabase
        """
        self.generic_test_write_data([int8oid], [int8_vdb_type_param], 'write_single_int8_type')

    
    def test_write_single_float4_type(self):
        """
        Tests whether tuples that only contain random float4 are successfully written by VictorDatabase
        """
        self.generic_test_write_data([float4oid], [float4_vdb_type_param], 'write_single_float4_type')

    
    def test_write_single_float8_type(self):
        """
        Tests whether tuples that only contain random float8 are successfully written by VictorDatabase
        """
        self.generic_test_write_data([float8oid], [float8_vdb_type_param], 'write_single_float8_type')

    
    def test_write_single_bool_array_type(self):
        """
        Tests whether tuples that only contain random bool array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([boolarrayoid], [bool_array_vdb_type_param], 'write_single_bool_array_type')

    
    def test_write_single_int2_array_type(self):
        """
        Tests whether tuples that only contain random int2 array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([int2arrayoid], [int2_array_vdb_type_param], 'write_single_int2_array_type')

    
    def test_write_single_int4_array_type(self):
        """
        Tests whether tuples that only contain random int4 array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([int4arrayoid], [int4_array_vdb_type_param], 'write_single_int4_array_type')

    
    def test_write_single_int8_array_type(self):
        """
        Tests whether tuples that only contain random int8 array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([int8arrayoid], [int8_array_vdb_type_param], 'write_single_int8_array_type')

    
    def test_write_single_float4_array_type(self):
        """
        Tests whether tuples that only contain random float4 array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([float4arrayoid], [float4_array_vdb_type_param], 'write_single_float4_array_type')

    
    def test_write_single_float8_array_type(self):
        """
        Tests whether tuples that only contain random float8 array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([float8arrayoid], [float8_array_vdb_type_param], 'write_single_float8_array_type')

    
    def test_write_single_float8_2d_array_type(self):
        """
        Tests whether tuples that only contain random float8 2d array are successfully written by VictorDatabase
        """
        self.generic_test_write_data([float82darrayoid], [float8_2d_array_vdb_type_param], 'write_single_float8_2d_array_type')

    
    def test_tuples_with_non_array_data(self):
        """
        This method tests write of tuples with possible non-array data type combinations.
        Since there are 7 non-array data types, all combinations produce 127 different schema.
        """
        test_tuples_with_non_array_data(self.generic_test_write_data, limit_test, with_sql)

    
    def test_tuples_with_array_data(self):
        """
        This method tests write of tuples with possible array(1d or 2d) data type combinations.
        Since there are 7 array data types, all combinations produce 127 different schema.
        """
        test_tuples_with_array_data(self.generic_test_write_data, limit_test, with_sql)

    
    def test_tuples_with_data(self):
        """
        This method tests write of tuples with possible non-array and array(1d or 2d) data type combinations.
        Since there are 14 data types, all combinations should produce 2 ** 14 - 1 different schema.
        """
        test_tuples_with_data(self.generic_test_write_data, limit_test, with_sql)
